import pytest
from unittest.mock import patch
from pipa.service.gengerate.export_pipashu_config_template import (
    query_filepath,
    generate_pipashu_template,
    generate_upload_template,
    CONFIG_TEMPLATE,
    UPLOAD_TEMPLATE,
)


def test_query_filepath_default(monkeypatch):
    mock_text = patch("questionary.text").start()
    mock_text.return_value.ask.return_value = ""
    monkeypatch.setattr("questionary.text", mock_text)
    assert query_filepath() == "./"
    patch.stopall()


def test_query_filepath_input(monkeypatch):
    mock_text = patch("questionary.text").start()
    mock_text.return_value.ask.return_value = "/path/to/config"
    monkeypatch.setattr("questionary.text", mock_text)
    assert query_filepath() == "/path/to/config"
    patch.stopall()


def test_generate_pipashu_template(monkeypatch):
    mock_generate_template = patch(
        "pipa.service.gengerate.export_pipashu_config_template.generate_template"
    ).start()
    mock_generate_template.return_value = len(CONFIG_TEMPLATE)
    monkeypatch.setattr(
        "pipa.service.gengerate.export_pipashu_config_template.generate_template",
        mock_generate_template,
    )
    result = generate_pipashu_template()
    mock_generate_template.assert_called_once_with(
        CONFIG_TEMPLATE, "config-pipa-shu.yaml"
    )
    assert result == len(CONFIG_TEMPLATE)
    patch.stopall()


def test_generate_upload_template(monkeypatch):
    filename = "config-upload.yaml"
    mock_generate_template = patch(
        "pipa.service.gengerate.export_pipashu_config_template.generate_template"
    ).start()
    mock_generate_template.return_value = len(UPLOAD_TEMPLATE)
    monkeypatch.setattr(
        "pipa.service.gengerate.export_pipashu_config_template.generate_template",
        mock_generate_template,
    )
    result = generate_upload_template(filename)
    mock_generate_template.assert_called_once_with(UPLOAD_TEMPLATE, filename)
    assert result == len(UPLOAD_TEMPLATE)
    patch.stopall()


if __name__ == "__main__":
    pytest.main([__file__])



import pandas as pd
from pipa.common.logger import logger
from pipa.common.utils import get_timestamp
from pipa.common.config import OUTPUT_DIR


def export_dataframe_to_csv(filepath=None):
    """
    Decorator function that exports a pandas DataFrame to a CSV file.

    Args:
        filepath (str, optional): The output file path. If not provided, a default path will be used.

    Returns:
        The decorated function.
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            if isinstance(result, pd.DataFrame):
                output_filepath = (
                    filepath or f"{OUTPUT_DIR}/export_{get_timestamp()}.csv"
                )
                logger.info(f"Exporting DataFrame to {output_filepath}")
                result.to_csv(output_filepath, index=False)
            return result

        return wrapper

    return decorator



import fire
from pipa.service.gengerate.all import quest_summary as generate_sh
from pipa.service.export_sys_config import run_export_config_script
from pipa.service.upload import main as pipa_upload
from pipa.service.dump import dump as pipa_dump
from pipa.common.utils import handle_user_cancelled
from pipa.__about__ import __version__
from rich import print


class PipaCLI:
    """
    The PipaCLI represents the command-line interface for PIPA (Platform Integrated Performance Analytics).
    It provides methods for generating performance collection scripts, exporting system configuration,
    uploading performance data, and displaying help and version information.
    Developed by: SPAIL, ZJU https://github.com/ZJU-SPAIL

    Usage:
      pipa generate
      pipa export
      pipa upload
      pipa dump
      pipa version
      pipa help

    Options:
      generate  Generate the performance collection scripts
      export    Export system configuration
      upload    Upload the performance data to PIPAD server
      dump      Dump PIPASHU overview data to a file
      version   Show the version of PIPA
      help      Show this help message and exit
    """

    @handle_user_cancelled
    def generate(self, config_path: str = None):
        # Generate the performance collection scripts
        generate_sh(config_path)

    @handle_user_cancelled
    def export(self):
        # Export system configuration
        run_export_config_script()

    @handle_user_cancelled
    def upload(self, config_path: str = None, verbose: bool = False):
        # Upload the performance data
        pipa_upload(config_path, verbose)

    def dump(self, output_path: str, config_path: str = None, verbose: bool = False):
        # Dump PIPASHU overview data to a file
        pipa_dump(output_path, config_path, verbose)

    def help(self):
        # Show this help message and exit
        print("PIPA (Platform Integrated Performance Analytics)")
        print("Developed by: SPAIL, ZJU https://github.com/ZJU-SPAIL")
        print("Usage:")
        print("  pipa generate")
        print("  pipa export")
        print("  pipa upload")
        print("  pipa dump")
        print("  pipa version")
        print("  pipa help")
        print("Options:")
        print("  generate  Generate the performance collection scripts")
        print("  export    Export system configuration")
        print("  upload    Upload the performance data to PIPAD server")
        print("  dump      Dump PIPASHU overview data to a file")
        print("  version   Show the version of PIPA")
        print("  help      Show this help message and exit")

    def version(self):
        # Show the version of PIPA
        print(f"PIPA (Platform Integrated Performance Analytics) version {__version__}")
        print("Developed by: SPAIL, ZJU https://github.com/ZJU-SPAIL")
        print("All rights reserved.")
        print("Licensed under the MIT License")
        print("https://github.com/ZJU-SPAIL/pipa")


def main():
    fire.Fire(PipaCLI)


if __name__ == "__main__":
    main()



from .upload import load
import yaml


def format(d: dict):
    """
    Formats the values in the given dictionary as floats.

    Args:
        d (dict): The dictionary containing the values to be formatted.

    Returns:
        dict: The dictionary with the values formatted as floats.
            - CPI (float): The cycles per instruction.
            - run_time (float): The run time of the workload.
            - cycles (float): The number of cycles.
            - instructions (float): The number of instructions.
            - path_length (float): The path length.
            - instructions_per_second (float): The instructions per second.
            - cycles_per_second (float): The cycles per second.
            - cycles_per_requests (float): The cycles per request.
            - throughput (float): The throughput.
    """
    d["CPI"] = float(d["CPI"])
    d["run_time"] = float(d["run_time"])
    d["cycles"] = float(d["cycles"])
    d["instructions"] = float(d["instructions"])
    d["path_length"] = float(d["path_length"])
    d["instructions_per_second"] = float(d["instructions_per_second"])
    d["cycles_per_second"] = float(d["cycles_per_second"])
    d["cycles_per_requests"] = float(d["cycles_per_requests"])
    d["throughput"] = float(d["throughput"])
    return d


def dump(output_path: str, config_path: str = None, verbose: bool = False):
    """
    Dump the data to a YAML file.

    Args:
        output_path (str): The path to the output file.
        config_path (str, optional): The path to the configuration file. Defaults to None.
        verbose (bool, optional): Whether to display verbose output. Defaults to False.

    Returns:
        dict: The dumped data.
    """
    data = load(config_path, verbose)
    with open(output_path, "w") as file:
        yaml.dump(data, file)
    return data


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Dump PIPASHU overview data to a file")
    parser.add_argument("-o", "--output_path", help="Path to the output file")
    parser.add_argument(
        "-c", "--config", help="Path to the configuration file", default=None
    )
    parser.add_argument("-v", "--verbose", help="Verbose mode", action="store_true")
    args = parser.parse_args()
    dump(args.output_path, args.config, args.verbose)



"""
pipad_server.py

start pipa grpc server:

```python
from pipa.service.pipad.pipad_server import PIPADServer
server = PIPADServer(data_location=dlocation, port=port, address=address, database=database, table=table)
server.serve()
```

you can specify the grafana url and api, PIPAD will automatically add a new connection of SQLite with path: data_location/database

```python
server = PIPADServer(data_location=dlocation, port=port, address=address, database=database, table=table, grafana_api_k=xxx, grafana_url=xxx)
```
"""

from concurrent import futures
from typing import Optional, Any
from pipa.common.logger import logger, stream_handler
from datetime import datetime
import time
import os
import requests
import json
import logging
import sqlite3
import grpc
import argparse
from . import pipad_pb2 as pipadlib
from . import pipad_pb2_grpc as pipadgrpc


def type_proto_to_sqlite(proto: type) -> str:
    """
    Convert a request params' type to its corresponding SQLite type.

    Args:
        proto (type): The request params' type to be converted.

    Returns:
        str: The SQLite type corresponding to proto.

    Raises:
        NotImplementedError: If the request params's type has no corresbonding type in SQLite.

    """
    if proto is int:
        return "INTEGER"
    elif proto is str:
        return "TEXT"
    elif proto is float:
        return "REAL"
    else:
        raise NotImplementedError(f"{proto}'s type can't be converted to sqlite type")


def value_to_sqlite_str(v: Any) -> str:
    """
    A function that converts a value got from request to a SQLite value str representation.

    Args:
        v (Any): The value to be converted.

    Returns:
        str: The SQLite str representation of the value.
    """
    t = type(v)
    if t is str:
        v = v.replace("'", "''")
        return f"'{v}'"
    elif t is int or t is float:
        return f"{v}"
    else:
        raise NotImplementedError(f"{v} has type {t}, can't be converted to sqlite str")


class PIPADServer:
    class PIPAService(pipadgrpc.PIPADServicer):
        def __init__(self, outer: "PIPADServer") -> None:
            """
            Initializes a new instance of the PIPADServer.PIPAService class.

            Args:
                outer ('PIPADServer'): The outer object that this instance belongs to. Should be PIPADServer.

            Returns:
                None
            """
            self._outer = outer
            super().__init__()

        def Deploy(
            self, request: pipadlib.DeployRequest, context
        ) -> pipadlib.DeployResp:
            """
            Deploy api, send your data and store it in sqlite file

            Args:
                request (pipadlib.DeployRequest): Your request, contains performance metrics

            Returns:
                pipadlib.DeployResp: Response to client
            """
            upload_time = time.time()
            kvpairs = []
            logger.debug(f"Received request: {request}")
            params = []
            ks = []
            vs = []
            for i, param in enumerate(request.ListFields()):
                name = param[0].name
                value = param[1]
                try:
                    t = type_proto_to_sqlite(type(value))
                except NotImplementedError as e:
                    logger.debug(f"{e}")
                    t = "TEXT"
                if i == 0:
                    record_prefix = f"{value}"
                try:
                    vtsqlites = value_to_sqlite_str(value)
                except NotImplementedError as e:
                    logger.debug(f"{e}")
                    vtsqlites = f"'{value}'"
                kvpairs.append(f"{name}{vtsqlites}")
                params.append(f"{name} {t}")
                ks.append(f"{name}")
                vs.append(f"{vtsqlites}")
            kvpairs.append(f"upload_time{upload_time}")
            hasht = f"'{hash(tuple(kvpairs))}'"
            upload_datetime = datetime.fromtimestamp(upload_time).strftime(
                r"%Y-%m-%d %H:%M:%S"
            )
            record = f"{record_prefix}-{upload_datetime}"
            kvs = ",".join(params)
            kss = ",".join(ks)
            vss = ",".join(vs)
            create_table_comm = f"CREATE TABLE IF NOT EXISTS {self._outer._table} (hash TEXT PRIMARY KEY,upload_time INTEGER,record TEXT,{kvs})"
            insert_table_comm = f"INSERT INTO {self._outer._table} (hash,upload_time,record,{kss}) VALUES ({hasht},{upload_time},{value_to_sqlite_str(record)},{vss})"
            logger.debug(f"Request to Table component: {kvs}")
            logger.debug(f"Request's Keys: {kss}")
            logger.debug(f"Request's Values: {vss}")
            logger.debug(f"Create table SQL: {create_table_comm}")
            logger.debug(f"Insert value SQL: {insert_table_comm}")
            try:
                with sqlite3.connect(self._outer._database_loc) as conn:
                    cursor = conn.cursor()
                    cursor.execute(create_table_comm)
                    cursor.execute(insert_table_comm)
                    conn.commit()
            except sqlite3.Error as e:
                logger.debug(f"Database Error: {e}")
                return pipadlib.DeployResp(
                    message=f"deploy failed. Contact admin. {e}", status_code=500
                )
            return pipadlib.DeployResp(
                message="deploy success",
                username=request.username,
                hash=hasht,
                time=f"{upload_time}",
                upload_datetime=upload_datetime,
                status_code=200,
            )

    def __init__(
        self,
        data_location: Optional[str] = "./",
        port: Optional[int] = 5051,
        address: Optional[str] = "[::]",
        database: str = "example.db",
        table: str = "example",
        grafana_api_k: Optional[str] = None,
        grafana_url: Optional[str] = None,
        grafana_path: Optional[str] = None,
    ) -> None:
        """
        Initialize the PIPADServer.

        Args:
            data_location (str, optional): The location to store the database files. Defaults to "./".
            port (int, optional): The port to bind to. Defaults to 5051.
            address (str, optional): The IP address to bind to. Defaults to "[::]".
            database (str, optional): The name of the database to store the results. Defaults to "example.db".
            table (str, optional): The name of the table in the database to store the results. Defaults to "example".
            grafana_api_k (str, optional): The API key for Grafana. Required if grafana_url is provided.
            grafana_url (str, optional): The URL for Grafana. Required if grafana_api_k is provided.
            grafana_path (str, optional): Replace the exact path name in grafana's database connection settings. Usefule for service in container.
        """
        self._dlocation = data_location
        self._port = port
        self._address = address
        self._database = database
        self._table = table
        self._database_loc = os.path.join(data_location, database)
        try:
            with sqlite3.connect(self._database_loc) as conn:
                logger.info(f"Use SQLite3: {sqlite3.sqlite_version}")
        except sqlite3.Error as e:
            logger.debug(
                f"Encounting problems when creating {self._database_loc}: {e}."
            )
            return
        if grafana_api_k is not None and grafana_url is not None:
            if grafana_path is not None:
                exact_path = f"{grafana_path}/{database}"
            else:
                exact_path = self._database_loc
            self._grafana = {"api_key": grafana_api_k, "url": grafana_url}
            new_conn = {
                "name": database,
                "type": "frser-sqlite-datasource",
                "access": "proxy",
                "jsonData": {"path": exact_path},
            }
            headers = {
                "Content-Type": "application/json",
                "Authorization": f"Bearer {grafana_api_k}",
            }

            response = requests.post(
                f"{grafana_url}/api/datasources",
                headers=headers,
                data=json.dumps(new_conn),
            )

            if response.status_code == 200:
                logger.info(f"New connection {self._database_loc} added successfully")
            elif response.status_code == 409:
                logger.debug(f"connection {self._database_loc} exists")
            else:
                logger.error(
                    f"Failed to add new connection {self._database_loc}: {response.status_code}"
                )
                logger.error(response.json())

    def serve(self) -> None:
        """
        Initiate the server, add the PIPAD service to the server, start the server, and wait for termination.
        """
        binding = f"{self._address}:{self._port}"
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
        pipadgrpc.add_PIPADServicer_to_server(self.PIPAService(self), server)
        server.add_insecure_port(binding)
        server.start()
        logger.info(f"Server started, listening on {binding}")
        try:
            server.wait_for_termination()
        except KeyboardInterrupt as e:
            logger.info("Server stop")
            server.stop(5)


def main():
    logger.setLevel(level=logging.DEBUG)
    stream_handler.setLevel(level=logging.DEBUG)
    argp = argparse.ArgumentParser()
    argp.add_argument(
        "-a", "--address", type=str, default="127.0.0.1", help="Specify address binding"
    )
    argp.add_argument(
        "-p", "--port", type=int, default=50051, help="Specify metrics port"
    )
    argp.add_argument(
        "-d",
        "--database",
        type=str,
        default="example.db",
        help="Specify which database to store data",
    )
    argp.add_argument(
        "-t", "--table", type=str, default="example", help="Specify table name"
    )
    argp.add_argument(
        "-l", "--data-location", type=str, default="./", help="Specify data location"
    )
    argp.add_argument(
        "--grafana-key",
        type=str,
        default=None,
        help="Specify grafana api key for connection",
    )
    argp.add_argument(
        "--grafana-url", type=str, default=None, help="Specify grafana url"
    )
    argp.add_argument(
        "--grafana-path",
        type=str,
        default=None,
        help="Specify grafana's exact data store path mapping with data-location",
    )
    args = argp.parse_args()
    address = getattr(args, "address")
    port = getattr(args, "port")
    database = getattr(args, "database")
    table = getattr(args, "table")
    dlocation = getattr(args, "data_location")
    gkey = getattr(args, "grafana_key")
    gurl = getattr(args, "grafana_url")
    gpath = getattr(args, "grafana_path")
    server = PIPADServer(
        data_location=dlocation,
        port=port,
        address=address,
        database=database,
        table=table,
        grafana_api_k=gkey,
        grafana_url=gurl,
        grafana_path=gpath,
    )
    server.serve()


if __name__ == "__main__":
    main()



import pytest
from unittest.mock import patch
from pipa.service.gengerate.parse_pipashu_config import build_command, quest, main


def test_build_command_without_taskset():
    command = build_command(False, "0-7", "echo test")
    assert command == "echo test"


def test_build_command_invalid_core_range():
    with pytest.raises(ValueError):
        build_command(True, "invalid_range", "echo test")


@patch("questionary.text")
def test_quest(mock_text):
    mock_text.return_value.ask.return_value = "./test-config.yaml"
    result = quest()
    assert result == "./test-config.yaml"


@patch("pipa.service.gengerate.parse_pipashu_config.quest")
@patch("pipa.service.gengerate.parse_pipashu_config.build")
def test_main_with_config_path(mock_build, mock_quest):
    config_path = "test-config.yaml"
    main(config_path)
    mock_build.assert_called_once_with(config_path)
    mock_quest.assert_not_called()


@patch("pipa.service.gengerate.parse_pipashu_config.quest")
@patch("pipa.service.gengerate.parse_pipashu_config.build")
def test_main_without_config_path(mock_build, mock_quest):
    mock_quest.return_value = "test-config.yaml"
    main()
    mock_quest.assert_called_once()
    mock_build.assert_called_once_with("test-config.yaml")


if __name__ == "__main__":
    pytest.main([__file__])



import questionary

from .run_by_pipa import main as run_by_pipa
from .run_by_user import main as run_by_user
from .export_pipashu_config_template import generate_pipashu_template
from .parse_pipashu_config import main as parse_pipashu_config
from .export_pipashu_config_template import generate_upload_template


def quest_summary(config_path: str = None):
    """
    Displays a menu to the user and performs different actions based on their selection.

    Args:
        config_path (str, optional): Path to the configuration file. Defaults to None.

    Returns:
        None
    """

    # TODO make these code more elegant
    if config_path:
        return parse_pipashu_config(config_path)

    how_to_run = questionary.select(
        "Please select the way of workload you want to run.",
        choices=[
            "Build scripts that collect global performance data.",
            "Build a script that collects performance data and start the workload by perf.",
            "Generate a configuration template configuration of PIPA-SHU.",
            "Build scripts based on the configuration file of PIPA-SHU.",
            "Generate a configuration template configuration of pipa-upload.",
            "Exit.",
        ],
    ).ask()

    if (
        how_to_run
        == "Build a script that collects performance data and start the workload by perf."
    ):
        run_by_pipa()
    elif how_to_run == "Build scripts that collect global performance data.":
        run_by_user()
    elif how_to_run == "Generate a configuration template configuration of PIPA-SHU.":
        generate_pipashu_template()
    elif how_to_run == "Build scripts based on the configuration file of PIPA-SHU.":
        parse_pipashu_config()
    elif (
        how_to_run == "Generate a configuration template configuration of pipa-upload."
    ):
        generate_upload_template()
    else:
        exit(0)



import subprocess
from pipa.common.logger import logger


def run_command(command: str, cwd: str = "./", log: bool = False):
    """
    Run a command in the specified directory.

    Args:
        command (str): The command to be executed.
        cwd (str, optional): The current working directory where the command will be executed. Defaults to "./".
        log (bool, optional): Flag indicating whether to log the command output. Defaults to False.

    Returns:
        str: The output of the command.

    Raises:
        Exception: If the command execution fails.
    """
    result = subprocess.run(
        command,
        shell=True,
        close_fds=True,
        cwd=cwd,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    suc = result.returncode == 0
    out = result.stdout.decode("UTF-8", "strict").strip()
    error = result.stderr.decode("UTF-8", "strict").strip()

    if not suc:
        logger.warning("'{}' return code = {}".format(command, result.returncode))
        raise Exception(error)
    else:
        if log:
            logger.info("{}: {}".format(command, out))
        else:
            logger.debug("{}: {}".format(command, out))
        return out



from __future__ import print_function

import sys
import struct
import sqlite3
import datetime

# An example of using this script with Intel PT:
#
# 	$ perf record -e intel_pt//u ls
# 	$ perf script -s src/pipa/service/db/load_perf_data_sqlite.py pt_example branches calls
# 	2017-07-31 14:26:07.326913 Creating database...
# 	2017-07-31 14:26:07.538097 Writing records...
# 	2017-07-31 14:26:09.889292 Adding indexes
# 	2017-07-31 14:26:09.958746 Done
#
# To browse the database, sqlite3 can be used e.g.
#
# 	$ sqlite3 pt_example
# 	sqlite> .header on
# 	sqlite> select * from samples_view where id < 10;
# 	sqlite> .mode column
# 	sqlite> select * from samples_view where id < 10;
# 	sqlite> .tables
# 	sqlite> .schema samples_view
# 	sqlite> .quit
#
# An example of using the database is provided by the script
# exported-sql-viewer.py.  Refer to that script for details.
#
# The database structure is practically the same as created by the script
# export-to-postgresql.py. Refer to that script for details.  A notable
# difference is  the 'transaction' column of the 'samples' table which is
# renamed 'transaction_' in sqlite because 'transaction' is a reserved word.

perf_db_export_mode = True
perf_db_export_calls = False
perf_db_export_callchains = False


def printerr(*args, **keyword_args):
    print(*args, file=sys.stderr, **keyword_args)


def printdate(*args, **kw_args):
    print(datetime.datetime.today(), *args, sep=" ", **kw_args)


def usage():
    printerr(
        "Usage is: export-to-sqlite.py <database name> [<columns>] [<calls>] [<callchains>] [<pyside-version-1>]"
    )
    printerr("where:  columns            'all' or 'branches'")
    printerr("        calls              'calls' => create calls and call_paths table")
    printerr("        callchains         'callchains' => create call_paths table")
    printerr("        pyside-version-1   'pyside-version-1' => use pyside version 1")
    raise Exception("Too few or bad arguments")


if len(sys.argv) < 2:
    usage()

dbname = sys.argv[1]

if len(sys.argv) >= 3:
    columns = sys.argv[2]
else:
    columns = "all"

if columns not in ("all", "branches"):
    usage()

branches = columns == "branches"

for i in range(3, len(sys.argv)):
    if sys.argv[i] == "calls":
        perf_db_export_calls = True
    elif sys.argv[i] == "callchains":
        perf_db_export_callchains = True
    elif sys.argv[i] == "pyside-version-1":
        pass
    else:
        usage()


def do_query(cursor, query, params=()):
    try:
        cursor.execute(query, params)
    except sqlite3.Error as e:
        raise Exception(f"Query failed: {e}")


def do_query_(cursor, query, params=()):
    try:
        cursor.execute(query, params)
    except sqlite3.Error as e:
        raise Exception(f"Query failed: {e}")


printdate("Creating database ...")

db_exists = False
try:
    f = open(dbname)
    f.close()
    db_exists = True
except Exception as e:
    pass

if db_exists:
    raise Exception(dbname + " already exists")

conn = sqlite3.connect(dbname)
cursor = conn.cursor()

do_query(cursor, "PRAGMA journal_mode = OFF")
do_query(cursor, "BEGIN TRANSACTION")

do_query(
    cursor,
    "CREATE TABLE selected_events ("
    "id		integer		NOT NULL	PRIMARY KEY,"
    "name		varchar(80))",
)
do_query(
    cursor,
    "CREATE TABLE machines ("
    "id		integer		NOT NULL	PRIMARY KEY,"
    "pid		integer,"
    "root_dir 	varchar(4096))",
)
do_query(
    cursor,
    "CREATE TABLE threads ("
    "id		integer		NOT NULL	PRIMARY KEY,"
    "machine_id	bigint,"
    "process_id	bigint,"
    "pid		integer,"
    "tid		integer)",
)
do_query(
    cursor,
    "CREATE TABLE comms ("
    "id		integer		NOT NULL	PRIMARY KEY,"
    "comm		varchar(16),"
    "c_thread_id	bigint,"
    "c_time		bigint,"
    "exec_flag	boolean)",
)
do_query(
    cursor,
    "CREATE TABLE comm_threads ("
    "id		integer		NOT NULL	PRIMARY KEY,"
    "comm_id	bigint,"
    "thread_id	bigint)",
)
do_query(
    cursor,
    "CREATE TABLE dsos ("
    "id		integer		NOT NULL	PRIMARY KEY,"
    "machine_id	bigint,"
    "short_name	varchar(256),"
    "long_name	varchar(4096),"
    "build_id	varchar(64))",
)
do_query(
    cursor,
    "CREATE TABLE symbols ("
    "id		integer		NOT NULL	PRIMARY KEY,"
    "dso_id		bigint,"
    "sym_start	bigint,"
    "sym_end	bigint,"
    "binding	integer,"
    "name		varchar(2048))",
)
do_query(
    cursor,
    "CREATE TABLE branch_types ("
    "id		integer		NOT NULL	PRIMARY KEY,"
    "name		varchar(80))",
)

if branches:
    do_query(
        cursor,
        "CREATE TABLE samples ("
        "id		integer		NOT NULL	PRIMARY KEY,"
        "evsel_id	bigint,"
        "machine_id	bigint,"
        "thread_id	bigint,"
        "comm_id	bigint,"
        "dso_id		bigint,"
        "symbol_id	bigint,"
        "sym_offset	bigint,"
        "ip		bigint,"
        "time		bigint,"
        "cpu		integer,"
        "to_dso_id	bigint,"
        "to_symbol_id	bigint,"
        "to_sym_offset	bigint,"
        "to_ip		bigint,"
        "branch_type	integer,"
        "in_tx		boolean,"
        "call_path_id	bigint,"
        "insn_count	bigint,"
        "cyc_count	bigint,"
        "flags		integer)",
    )
else:
    do_query(
        cursor,
        "CREATE TABLE samples ("
        "id		integer		NOT NULL	PRIMARY KEY,"
        "evsel_id	bigint,"
        "machine_id	bigint,"
        "thread_id	bigint,"
        "comm_id	bigint,"
        "dso_id		bigint,"
        "symbol_id	bigint,"
        "sym_offset	bigint,"
        "ip		bigint,"
        "time		bigint,"
        "cpu		integer,"
        "to_dso_id	bigint,"
        "to_symbol_id	bigint,"
        "to_sym_offset	bigint,"
        "to_ip		bigint,"
        "period		bigint,"
        "weight		bigint,"
        "transaction_	bigint,"
        "data_src	bigint,"
        "branch_type	integer,"
        "in_tx		boolean,"
        "call_path_id	bigint,"
        "insn_count	bigint,"
        "cyc_count	bigint,"
        "flags		integer)",
    )

if perf_db_export_calls or perf_db_export_callchains:
    do_query(
        cursor,
        "CREATE TABLE call_paths ("
        "id		integer		NOT NULL	PRIMARY KEY,"
        "parent_id	bigint,"
        "symbol_id	bigint,"
        "ip		bigint)",
    )
if perf_db_export_calls:
    do_query(
        cursor,
        "CREATE TABLE calls ("
        "id		integer		NOT NULL	PRIMARY KEY,"
        "thread_id	bigint,"
        "comm_id	bigint,"
        "call_path_id	bigint,"
        "call_time	bigint,"
        "return_time	bigint,"
        "branch_count	bigint,"
        "call_id	bigint,"
        "return_id	bigint,"
        "parent_call_path_id	bigint,"
        "flags		integer,"
        "parent_id	bigint,"
        "insn_count	bigint,"
        "cyc_count	bigint)",
    )

do_query(
    cursor,
    "CREATE TABLE ptwrite ("
    "id		integer		NOT NULL	PRIMARY KEY,"
    "payload	bigint,"
    "exact_ip	integer)",
)

do_query(
    cursor,
    "CREATE TABLE cbr ("
    "id		integer		NOT NULL	PRIMARY KEY,"
    "cbr		integer,"
    "mhz		integer,"
    "percent	integer)",
)

do_query(
    cursor,
    "CREATE TABLE mwait ("
    "id		integer		NOT NULL	PRIMARY KEY,"
    "hints		integer,"
    "extensions	integer)",
)

do_query(
    cursor,
    "CREATE TABLE pwre ("
    "id		integer		NOT NULL	PRIMARY KEY,"
    "cstate		integer,"
    "subcstate	integer,"
    "hw		integer)",
)

do_query(
    cursor,
    "CREATE TABLE exstop (" "id		integer		NOT NULL	PRIMARY KEY," "exact_ip	integer)",
)

do_query(
    cursor,
    "CREATE TABLE pwrx ("
    "id		integer		NOT NULL	PRIMARY KEY,"
    "deepest_cstate	integer,"
    "last_cstate	integer,"
    "wake_reason	integer)",
)

do_query(
    cursor,
    "CREATE TABLE context_switches ("
    "id		integer		NOT NULL	PRIMARY KEY,"
    "machine_id	bigint,"
    "time		bigint,"
    "cpu		integer,"
    "thread_out_id	bigint,"
    "comm_out_id	bigint,"
    "thread_in_id	bigint,"
    "comm_in_id	bigint,"
    "flags		integer)",
)

# printf was added to sqlite in version 3.8.3
sqlite_has_printf = False
try:
    do_query(cursor, 'SELECT printf("") FROM machines')
    sqlite_has_printf = True
except sqlite3.Error:
    pass


def emit_to_hex(x):
    # if sqlite_has_printf:
    #     return 'printf("%x", ' + x + ")"
    # else:
    return x


do_query(
    cursor,
    "CREATE VIEW machines_view AS "
    "SELECT "
    "id,"
    "pid,"
    "root_dir,"
    "CASE WHEN id=0 THEN 'unknown' WHEN pid=-1 THEN 'host' ELSE 'guest' END AS host_or_guest"
    " FROM machines",
)

do_query(
    cursor,
    "CREATE VIEW dsos_view AS "
    "SELECT "
    "id,"
    "machine_id,"
    "(SELECT host_or_guest FROM machines_view WHERE id = machine_id) AS host_or_guest,"
    "short_name,"
    "long_name,"
    "build_id"
    " FROM dsos",
)

do_query(
    cursor,
    "CREATE VIEW symbols_view AS "
    "SELECT "
    "id,"
    "name,"
    "(SELECT short_name FROM dsos WHERE id=dso_id) AS dso,"
    "dso_id,"
    "sym_start,"
    "sym_end,"
    "CASE WHEN binding=0 THEN 'local' WHEN binding=1 THEN 'global' ELSE 'weak' END AS binding"
    " FROM symbols",
)

do_query(
    cursor,
    "CREATE VIEW threads_view AS "
    "SELECT "
    "id,"
    "machine_id,"
    "(SELECT host_or_guest FROM machines_view WHERE id = machine_id) AS host_or_guest,"
    "process_id,"
    "pid,"
    "tid"
    " FROM threads",
)

do_query(
    cursor,
    "CREATE VIEW comm_threads_view AS "
    "SELECT "
    "comm_id,"
    "(SELECT comm FROM comms WHERE id = comm_id) AS command,"
    "thread_id,"
    "(SELECT pid FROM threads WHERE id = thread_id) AS pid,"
    "(SELECT tid FROM threads WHERE id = thread_id) AS tid"
    " FROM comm_threads",
)

if perf_db_export_calls or perf_db_export_callchains:
    do_query(
        cursor,
        "CREATE VIEW call_paths_view AS "
        "SELECT "
        "c.id," + emit_to_hex("c.ip") + " AS ip,"
        "c.symbol_id,"
        "(SELECT name FROM symbols WHERE id = c.symbol_id) AS symbol,"
        "(SELECT dso_id FROM symbols WHERE id = c.symbol_id) AS dso_id,"
        "(SELECT dso FROM symbols_view  WHERE id = c.symbol_id) AS dso_short_name,"
        "c.parent_id," + emit_to_hex("p.ip") + " AS parent_ip,"
        "p.symbol_id AS parent_symbol_id,"
        "(SELECT name FROM symbols WHERE id = p.symbol_id) AS parent_symbol,"
        "(SELECT dso_id FROM symbols WHERE id = p.symbol_id) AS parent_dso_id,"
        "(SELECT dso FROM symbols_view  WHERE id = p.symbol_id) AS parent_dso_short_name"
        " FROM call_paths c INNER JOIN call_paths p ON p.id = c.parent_id",
    )
if perf_db_export_calls:
    do_query(
        cursor,
        "CREATE VIEW calls_view AS "
        "SELECT "
        "calls.id,"
        "thread_id,"
        "(SELECT pid FROM threads WHERE id = thread_id) AS pid,"
        "(SELECT tid FROM threads WHERE id = thread_id) AS tid,"
        "(SELECT comm FROM comms WHERE id = comm_id) AS command,"
        "call_path_id," + emit_to_hex("ip") + " AS ip,"
        "symbol_id,"
        "(SELECT name FROM symbols WHERE id = symbol_id) AS symbol,"
        "call_time,"
        "return_time,"
        "return_time - call_time AS elapsed_time,"
        "branch_count,"
        "insn_count,"
        "cyc_count,"
        "CASE WHEN cyc_count=0 THEN CAST(0 AS FLOAT) ELSE ROUND(CAST(insn_count AS FLOAT) / cyc_count, 2) END AS IPC,"
        "call_id,"
        "return_id,"
        "CASE WHEN flags=0 THEN '' WHEN flags=1 THEN 'no call' WHEN flags=2 THEN 'no return' WHEN flags=3 THEN 'no call/return' WHEN flags=6 THEN 'jump' ELSE flags END AS flags,"
        "parent_call_path_id,"
        "calls.parent_id"
        " FROM calls INNER JOIN call_paths ON call_paths.id = call_path_id",
    )

do_query(
    cursor,
    "CREATE VIEW samples_view AS "
    "SELECT "
    "id,"
    "time,"
    "cpu,"
    "(SELECT pid FROM threads WHERE id = thread_id) AS pid,"
    "(SELECT tid FROM threads WHERE id = thread_id) AS tid,"
    "(SELECT comm FROM comms WHERE id = comm_id) AS command,"
    "(SELECT name FROM selected_events WHERE id = evsel_id) AS event,"
    + emit_to_hex("ip")
    + " AS ip_hex,"
    "(SELECT name FROM symbols WHERE id = symbol_id) AS symbol,"
    "sym_offset,"
    "(SELECT short_name FROM dsos WHERE id = dso_id) AS dso_short_name,"
    + emit_to_hex("to_ip")
    + " AS to_ip_hex,"
    "(SELECT name FROM symbols WHERE id = to_symbol_id) AS to_symbol,"
    "to_sym_offset,"
    "(SELECT short_name FROM dsos WHERE id = to_dso_id) AS to_dso_short_name,"
    "(SELECT name FROM branch_types WHERE id = branch_type) AS branch_type_name,"
    "in_tx,"
    "insn_count,"
    "cyc_count,"
    "CASE WHEN cyc_count=0 THEN CAST(0 AS FLOAT) ELSE ROUND(CAST(insn_count AS FLOAT) / cyc_count, 2) END AS IPC,"
    "flags"
    " FROM samples",
)

do_query(
    cursor,
    "CREATE VIEW ptwrite_view AS "
    "SELECT "
    "ptwrite.id,"
    "time,"
    "cpu," + emit_to_hex("payload") + " AS payload_hex,"
    "CASE WHEN exact_ip=0 THEN 'False' ELSE 'True' END AS exact_ip"
    " FROM ptwrite"
    " INNER JOIN samples ON samples.id = ptwrite.id",
)

do_query(
    cursor,
    "CREATE VIEW cbr_view AS "
    "SELECT "
    "cbr.id,"
    "time,"
    "cpu,"
    "cbr,"
    "mhz,"
    "percent"
    " FROM cbr"
    " INNER JOIN samples ON samples.id = cbr.id",
)

do_query(
    cursor,
    "CREATE VIEW mwait_view AS "
    "SELECT "
    "mwait.id,"
    "time,"
    "cpu,"
    + emit_to_hex("hints")
    + " AS hints_hex,"
    + emit_to_hex("extensions")
    + " AS extensions_hex"
    " FROM mwait"
    " INNER JOIN samples ON samples.id = mwait.id",
)

do_query(
    cursor,
    "CREATE VIEW pwre_view AS "
    "SELECT "
    "pwre.id,"
    "time,"
    "cpu,"
    "cstate,"
    "subcstate,"
    "CASE WHEN hw=0 THEN 'False' ELSE 'True' END AS hw"
    " FROM pwre"
    " INNER JOIN samples ON samples.id = pwre.id",
)

do_query(
    cursor,
    "CREATE VIEW exstop_view AS "
    "SELECT "
    "exstop.id,"
    "time,"
    "cpu,"
    "CASE WHEN exact_ip=0 THEN 'False' ELSE 'True' END AS exact_ip"
    " FROM exstop"
    " INNER JOIN samples ON samples.id = exstop.id",
)

do_query(
    cursor,
    "CREATE VIEW pwrx_view AS "
    "SELECT "
    "pwrx.id,"
    "time,"
    "cpu,"
    "deepest_cstate,"
    "last_cstate,"
    "CASE     WHEN wake_reason=1 THEN 'Interrupt'"
    " WHEN wake_reason=2 THEN 'Timer Deadline'"
    " WHEN wake_reason=4 THEN 'Monitored Address'"
    " WHEN wake_reason=8 THEN 'HW'"
    " ELSE wake_reason "
    "END AS wake_reason"
    " FROM pwrx"
    " INNER JOIN samples ON samples.id = pwrx.id",
)

do_query(
    cursor,
    "CREATE VIEW power_events_view AS "
    "SELECT "
    "samples.id,"
    "time,"
    "cpu,"
    "selected_events.name AS event,"
    "CASE WHEN selected_events.name='cbr' THEN (SELECT cbr FROM cbr WHERE cbr.id = samples.id) ELSE \"\" END AS cbr,"
    "CASE WHEN selected_events.name='cbr' THEN (SELECT mhz FROM cbr WHERE cbr.id = samples.id) ELSE \"\" END AS mhz,"
    "CASE WHEN selected_events.name='cbr' THEN (SELECT percent FROM cbr WHERE cbr.id = samples.id) ELSE \"\" END AS percent,"
    "CASE WHEN selected_events.name='mwait' THEN (SELECT "
    + emit_to_hex("hints")
    + ' FROM mwait WHERE mwait.id = samples.id) ELSE "" END AS hints_hex,'
    "CASE WHEN selected_events.name='mwait' THEN (SELECT "
    + emit_to_hex("extensions")
    + ' FROM mwait WHERE mwait.id = samples.id) ELSE "" END AS extensions_hex,'
    "CASE WHEN selected_events.name='pwre' THEN (SELECT cstate FROM pwre WHERE pwre.id = samples.id) ELSE \"\" END AS cstate,"
    "CASE WHEN selected_events.name='pwre' THEN (SELECT subcstate FROM pwre WHERE pwre.id = samples.id) ELSE \"\" END AS subcstate,"
    "CASE WHEN selected_events.name='pwre' THEN (SELECT hw FROM pwre WHERE pwre.id = samples.id) ELSE \"\" END AS hw,"
    "CASE WHEN selected_events.name='exstop' THEN (SELECT exact_ip FROM exstop WHERE exstop.id = samples.id) ELSE \"\" END AS exact_ip,"
    "CASE WHEN selected_events.name='pwrx' THEN (SELECT deepest_cstate FROM pwrx WHERE pwrx.id = samples.id) ELSE \"\" END AS deepest_cstate,"
    "CASE WHEN selected_events.name='pwrx' THEN (SELECT last_cstate FROM pwrx WHERE pwrx.id = samples.id) ELSE \"\" END AS last_cstate,"
    "CASE WHEN selected_events.name='pwrx' THEN (SELECT "
    "CASE     WHEN wake_reason=1 THEN 'Interrupt'"
    " WHEN wake_reason=2 THEN 'Timer Deadline'"
    " WHEN wake_reason=4 THEN 'Monitored Address'"
    " WHEN wake_reason=8 THEN 'HW'"
    " ELSE wake_reason "
    "END"
    ' FROM pwrx WHERE pwrx.id = samples.id) ELSE "" END AS wake_reason'
    " FROM samples"
    " INNER JOIN selected_events ON selected_events.id = evsel_id"
    " WHERE selected_events.name IN ('cbr','mwait','exstop','pwre','pwrx')",
)

do_query(
    cursor,
    "CREATE VIEW context_switches_view AS "
    "SELECT "
    "context_switches.id,"
    "context_switches.machine_id,"
    "context_switches.time,"
    "context_switches.cpu,"
    "th_out.pid AS pid_out,"
    "th_out.tid AS tid_out,"
    "comm_out.comm AS comm_out,"
    "th_in.pid AS pid_in,"
    "th_in.tid AS tid_in,"
    "comm_in.comm AS comm_in,"
    "CASE	  WHEN context_switches.flags = 0 THEN 'in'"
    " WHEN context_switches.flags = 1 THEN 'out'"
    " WHEN context_switches.flags = 3 THEN 'out preempt'"
    " ELSE context_switches.flags "
    "END AS flags"
    " FROM context_switches"
    " INNER JOIN threads AS th_out ON th_out.id   = context_switches.thread_out_id"
    " INNER JOIN threads AS th_in  ON th_in.id    = context_switches.thread_in_id"
    " INNER JOIN comms AS comm_out ON comm_out.id = context_switches.comm_out_id"
    " INNER JOIN comms AS comm_in  ON comm_in.id  = context_switches.comm_in_id",
)

do_query(cursor, "END TRANSACTION")
conn.commit()


printdate("Database setup completed.")


# 插入数据函数
def insert_selected_events(cursor, values):
    cursor.execute("INSERT INTO selected_events VALUES (?, ?)", values)


def insert_machines(cursor, values):
    cursor.execute("INSERT INTO machines VALUES (?, ?, ?)", values)


def insert_threads(cursor, values):
    cursor.execute("INSERT INTO threads VALUES (?, ?, ?, ?, ?)", values)


def insert_comms(cursor, values):
    cursor.execute("INSERT INTO comms VALUES (?, ?, ?, ?, ?)", values)


def insert_comm_threads(cursor, values):
    cursor.execute("INSERT INTO comm_threads VALUES (?, ?, ?)", values)


def insert_dsos(cursor, values):
    cursor.execute("INSERT INTO dsos VALUES (?, ?, ?, ?, ?)", values)


def insert_symbols(cursor, values):
    cursor.execute("INSERT INTO symbols VALUES (?, ?, ?, ?, ?, ?)", values)


def insert_branch_types(cursor, values):
    cursor.execute("INSERT INTO branch_types VALUES (?, ?)", values)


def insert_samples(cursor, values):
    if branches:
        cursor.execute(
            "INSERT INTO samples VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
            values,
        )
    else:
        cursor.execute(
            "INSERT INTO samples VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
            values,
        )


def insert_call_paths(cursor, values):
    cursor.execute("INSERT INTO call_paths VALUES (?, ?, ?, ?)", values)


def insert_calls(cursor, values):
    cursor.execute(
        "INSERT INTO calls VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)", values
    )


def trace_begin():
    printdate("Writing records...")
    do_query(cursor, "BEGIN TRANSACTION")

    # 插入初始数据
    insert_selected_events(cursor, (0, "unknown"))
    insert_machines(cursor, (0, 0, "unknown"))
    insert_threads(cursor, (0, 0, 0, -1, -1))
    insert_comms(cursor, (0, "unknown", 0, 0, 0))
    insert_dsos(cursor, (0, 0, "unknown", "unknown", ""))
    insert_symbols(cursor, (0, 0, 0, 0, 0, "unknown"))

    if branches:
        insert_samples(
            cursor, (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
        )
    else:
        insert_samples(
            cursor,
            (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
        )

    if perf_db_export_calls or perf_db_export_callchains:
        insert_call_paths(cursor, (0, 0, 0, 0))
        insert_calls(cursor, (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0))


unhandled_count = 0


def trace_end():
    do_query(cursor, "END TRANSACTION")
    conn.commit()

    printdate("Adding indexes")
    if perf_db_export_calls:
        do_query(cursor, "CREATE INDEX pcpid_idx ON calls (parent_call_path_id)")
        do_query(cursor, "CREATE INDEX pid_idx ON calls (parent_id)")
        do_query(cursor, "ALTER TABLE comms ADD COLUMN has_calls boolean")
        do_query(
            cursor,
            "UPDATE comms SET has_calls = 1 WHERE comms.id IN (SELECT DISTINCT comm_id FROM calls)",
        )

    printdate("Dropping unused tables")
    if is_table_empty("ptwrite"):
        drop("ptwrite")
    if (
        is_table_empty("mwait")
        and is_table_empty("pwre")
        and is_table_empty("exstop")
        and is_table_empty("pwrx")
    ):
        do_query(cursor, "DROP VIEW power_events_view")
        drop("mwait")
        drop("pwre")
        drop("exstop")
        drop("pwrx")
        if is_table_empty("cbr"):
            drop("cbr")
    if is_table_empty("context_switches"):
        drop("context_switches")

    if unhandled_count:
        printdate("Warning: ", unhandled_count, " unhandled events")
    printdate("Done")


def is_table_empty(table_name):
    do_query(cursor, f"SELECT * FROM {table_name} LIMIT 1")
    if cursor.fetchone():
        return False
    return True


def drop(table_name):
    do_query(cursor, f"DROP VIEW IF EXISTS {table_name}_view")
    do_query(cursor, f"DROP TABLE IF EXISTS {table_name}")


def trace_unhandled(event_name, context, event_fields_dict):
    global unhandled_count
    unhandled_count += 1


def sched__sched_switch(*x):
    pass


def bind_exec(cursor, query, values):
    cursor.execute(query, values)


def evsel_table(*x):
    bind_exec(cursor, "INSERT INTO selected_events VALUES (?, ?)", x)


def machine_table(*x):
    bind_exec(cursor, "INSERT INTO machines VALUES (?, ?, ?)", x)


def thread_table(*x):
    bind_exec(cursor, "INSERT INTO threads VALUES (?, ?, ?, ?, ?)", x)


def comm_table(*x):
    bind_exec(cursor, "INSERT INTO comms VALUES (?, ?, ?, ?, ?)", x)


def comm_thread_table(*x):
    bind_exec(cursor, "INSERT INTO comm_threads VALUES (?, ?, ?)", x)


def dso_table(*x):
    bind_exec(cursor, "INSERT INTO dsos VALUES (?, ?, ?, ?, ?)", x)


def symbol_table(*x):
    bind_exec(cursor, "INSERT INTO symbols VALUES (?, ?, ?, ?, ?, ?)", x)


def branch_type_table(*x):
    bind_exec(cursor, "INSERT INTO branch_types VALUES (?, ?)", x)


def sample_table(*x):
    if branches:
        values = x[:15] + x[19:25]
        cursor.execute(
            "INSERT INTO samples VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
            values,
        )
    else:
        cursor.execute(
            "INSERT INTO samples VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
            x[:25],
        )


def call_path_table(*x):
    bind_exec(cursor, "INSERT INTO call_paths VALUES (?, ?, ?, ?)", x)


def call_return_table(*x):
    bind_exec(
        cursor, "INSERT INTO calls VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)", x
    )


def ptwrite(id, raw_buf):
    data = struct.unpack_from("<IQ", raw_buf)
    flags = data[0]
    payload = data[1]
    exact_ip = flags & 1
    insert_values = (id, payload, exact_ip)
    bind_exec(cursor, "INSERT INTO ptwrite VALUES (?, ?, ?)", insert_values)


def cbr(id, raw_buf):
    data = struct.unpack_from("<BBBBII", raw_buf)
    cbr = data[0]
    MHz = (data[4] + 500) / 1000
    percent = ((cbr * 1000 / data[2]) + 5) / 10
    insert_values = (id, cbr, MHz, percent)
    bind_exec(cursor, "INSERT INTO cbr VALUES (?, ?, ?, ?)", insert_values)


def mwait(id, raw_buf):
    data = struct.unpack_from("<IQ", raw_buf)
    payload = data[1]
    hints = payload & 0xFF
    extensions = (payload >> 32) & 0x3
    insert_values = (id, hints, extensions)
    bind_exec(cursor, "INSERT INTO mwait VALUES (?, ?, ?)", insert_values)


def pwre(id, raw_buf):
    data = struct.unpack_from("<IQ", raw_buf)
    payload = data[1]
    hw = (payload >> 7) & 1
    cstate = (payload >> 12) & 0xF
    subcstate = (payload >> 8) & 0xF
    insert_values = (id, cstate, subcstate, hw)
    bind_exec(cursor, "INSERT INTO pwre VALUES (?, ?, ?, ?)", insert_values)


def exstop(id, raw_buf):
    data = struct.unpack_from("<I", raw_buf)
    flags = data[0]
    exact_ip = flags & 1
    insert_values = (id, exact_ip)
    bind_exec(cursor, "INSERT INTO exstop VALUES (?, ?)", insert_values)


def pwrx(id, raw_buf):
    data = struct.unpack_from("<IQ", raw_buf)
    payload = data[1]
    deepest_cstate = payload & 0xF
    last_cstate = (payload >> 4) & 0xF
    wake_reason = (payload >> 8) & 0xF
    insert_values = (id, deepest_cstate, last_cstate, wake_reason)
    bind_exec(cursor, "INSERT INTO pwrx VALUES (?, ?, ?, ?)", insert_values)


def synth_data(id, config, raw_buf, *x):
    if config == 0:
        ptwrite(id, raw_buf)
    elif config == 1:
        mwait(id, raw_buf)
    elif config == 2:
        pwre(id, raw_buf)
    elif config == 3:
        exstop(id, raw_buf)
    elif config == 4:
        pwrx(id, raw_buf)
    elif config == 5:
        cbr(id, raw_buf)


def context_switch_table(*x):
    bind_exec(
        cursor, "INSERT INTO context_switches VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)", x
    )



from pipa.service.gengerate.common import load_yaml_config
from pipa.service.gengerate.run_by_pipa import generate as generate_pipa
from pipa.service.gengerate.run_by_user import generate as generate_user
from pipa.common.hardware.cpu import get_cpu_cores

import questionary


def build_command(use_taskset: bool, core_range, command):
    """
    Builds a command with optional taskset and core range.

    Args:
        use_taskset (bool): Whether to use taskset for CPU affinity.
        core_range (str): The range of CPU cores to use.
        command (str): The command to be executed.

    Returns:
        str: The built command.

    Raises:
        ValueError: If the core range is not a valid range or contains non-digit characters.

    """
    if use_taskset:
        CORES_ALL = get_cpu_cores()
        if core_range.isdigit():
            core_list = core_range.strip()
        elif core_range.split("-").__len__() != 2:
            raise ValueError("Please input cores as a valid range, split by '-'.")
        else:
            left, right = core_range.split("-")

            left, right = left.strip(), right.strip()
            if not left.isdigit() or not right.isdigit():
                raise ValueError(
                    "Please input cores as a valid range, non-digit char detected."
                )
            left, right = int(left), int(right)
            if left < CORES_ALL[0] or right > CORES_ALL[-1] or left > right:
                raise ValueError("Please input cores as a valid range.")
            core_list = ",".join([str(i) for i in list(range(left, right + 1))])

        command = f"/usr/bin/taskset -c {core_list} {command}"
    return command


def quest():
    """
    Asks the user for the location of the configuration file of PIPA-SHU.

    Returns:
        str: The path to the configuration file.
    """
    config_yaml = questionary.text(
        "Where is the configuration file of PIPA-SHU?\n", "./config-pipa-shu.yaml"
    ).ask()
    return config_yaml


def build(path: str):
    """
    Builds the configuration based on the provided path.

    Args:
        path (str): The path to the configuration file.

    Returns:
        None
    """
    config = load_yaml_config(path)
    config["events_stat"] = ",".join(config["events_stat"])
    build_command(config["use_taskset"], config["core_range"], config["command"])
    if config["run_by_perf"]:
        generate_pipa(config)
    else:
        generate_user(config)


def main(config_path: str = None):
    """
    Main function for parsing Pipashu configuration.

    Args:
        config_path (str, optional): Path to the configuration file. If not provided, a prompt will be shown to enter the path.

    Returns:
        The result of the `build` function using the provided or prompted configuration path.
    """
    if config_path is None:
        config_path = quest()
    return build(config_path)


if __name__ == "__main__":
    main()



import datetime
import sys
from rich import print


def get_timestamp():
    """
    Returns the current timestamp in the format "YYYY-MM-DD-HH-MM-SS".

    Returns:
        str: The current timestamp.
    """
    return datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")


def handle_user_cancelled(func):
    """
    Decorator function that handles user cancellation by catching the KeyboardInterrupt exception.

    Args:
        func: The function to be decorated.

    Returns:
        The decorated function.

    Raises:
        None.
    """

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            # Print "Cancelled by user" and exit
            print("Cancelled by user")
            sys.exit(0)

    return wrapper



import questionary
from rich import print
from pipa.service.gengerate.common import quest_basic, CORES_ALL, write_title, opener
from pipa.service.export_sys_config import write_export_config_script
import os


def quest():
    """
    This function prompts the user with a series of questions to gather information for running a workload.

    Returns:
        A tuple containing the following information:
        - workspace: The workspace path
        - freq_record: Frequency record
        - events_record: Events record
        - freq_stat: Frequency statistics
        - events_stat: Events statistics
        - annotete: Whether to use perf-annotate (True or False)
        - command: The command of the workload
    """
    config = quest_basic()

    use_taskset = questionary.select(
        "Whether to use taskset?\n", choices=["Yes", "No"]
    ).ask()

    if use_taskset == "Yes":
        cores_input = questionary.text(
            f"Which cores do you want to use? (Default: {CORES_ALL[0]}-{CORES_ALL[-1]})\n",
            f"{CORES_ALL[0]}-{CORES_ALL[-1]}",
        ).ask()

        if cores_input.isdigit():
            core_list = cores_input.strip()
        elif cores_input.split("-").__len__() != 2:
            raise ("Please input cores as a valid range, split by '-'.")
        else:
            left, right = cores_input.split("-")

            left, right = left.strip(), right.strip()
            if not left.isdigit() or not right.isdigit():
                raise ("Please input cores as a valid range, non-digit char detected.")
            left, right = int(left), int(right)
            if left < CORES_ALL[0] or right > CORES_ALL[-1] or left > right:
                raise ("Please input cores as a valid range.")
            core_list = ",".join([str(i) for i in list(range(left, right + 1))])

    command = questionary.text("What's the command of workload?\n").ask()

    if command == "":
        print("Please input a command to run workload.")
        exit(1)

    if use_taskset == "Yes":
        config["command"] = f"/usr/bin/taskset -c {core_list} {command}"
    else:
        config["command"] = command
    return config


def generate(config):
    """
    Generate a shell script for collecting performance data which start workload by perf.

    Args:
        config (dict): Configuration dictionary containing the following keys:
            - workspace (str): The path to the workspace directory.
            - freq_record (str): The frequency for recording events.
            - events_record (str): The events to be recorded.
            - freq_stat (str): The frequency for collecting statistics.
            - events_stat (str): The events to be collected for statistics.
            - annotete (bool): Whether to annotate the performance data.
            - command (str): The command to be executed.

    Returns:
        None
    """
    workspace = config["workspace"]
    freq_record = config["freq_record"]
    events_record = config["events_record"]
    count_delta_stat = config["count_delta_stat"]
    events_stat = config["events_stat"]
    annotete = config["annotete"]
    command = config["command"]
    use_emon = config["use_emon"]
    if use_emon:
        mpp = config["MPP_HOME"]
    with open(workspace + "/pipa-run.sh", "w", opener=opener) as f:
        write_title(f)

        f.write("WORKSPACE=" + workspace + "\n")
        f.write("mkdir -p $WORKSPACE\n\n")

        f.write("ps -aux -ef --forest --sort=-%cpu > $WORKSPACE/ps.txt\n")

        f.write(
            f"perf record -e '{events_record}' -a -F {freq_record} -o $WORKSPACE/perf.data {command}\n"
        )
        f.write(
            "perf script -i $WORKSPACE/perf.data -I --header > $WORKSPACE/perf.script\n"
        )
        f.write(
            "perf report -i $WORKSPACE/perf.data -I --header > $WORKSPACE/perf.report\n\n"
        )

        f.write("sar -o $WORKSPACE/sar.dat 1 >/dev/null 2>&1 &\n")
        f.write("sar_pid=$!\n")

        if use_emon:
            f.write(
                f"emon -i {mpp}/emon_event_all.txt -v -f $WORKSPACE/emon_result.txt -t 0.1 -l 100000000 -c -experimental -w {command} &\n"
            )
        else:
            f.write(
                f"perf stat -e {events_stat} -C {CORES_ALL[0]}-{CORES_ALL[-1]} -A -x , -I {count_delta_stat} -o $WORKSPACE/perf-stat.csv {command}\n"
            )

        f.write("kill -9 $sar_pid\n")
        f.write("LC_ALL='C' sar -A -f $WORKSPACE/sar.dat >$WORKSPACE/sar.txt\n\n")

        if use_emon:
            f.write(
                f"python {mpp}/mpp/mpp.py -i $WORKSPACE/emon_result.txt -m {mpp}/metrics/icelake_server_2s_nda.xml -o ./ --thread-view"
            )

        if annotete:
            f.write(
                "perf annotate -i $WORKSPACE/perf.data > $WORKSPACE/perf.annotate\n\n"
            )

        write_export_config_script(f, os.path.join(workspace, "config"))

        f.write("echo 'Performance data collected successfully.'\n")

        print("Shell script generated successfully.")
        print("Please check the script in " + workspace + "/pipa-run.sh")
        print(
            "Note that you may need to modify the script to meet your requirements.",
            "The core list is generated according to the machine which runs this script now.",
            "pipa-run.sh is more suitable for reproducible workloads such as benchmark.",
        )


def main():
    generate(quest())


if __name__ == "__main__":
    main()



import pandas as pd
import re
import multiprocessing
import os
from pipa.common.logger import logger
from pipa.common.hardware.cpu import NUM_CORES_PHYSICAL
from pandarallel import pandarallel


class PerfScriptData:
    def __init__(
        self, parsed_script_path: str, threads_num=min(12, NUM_CORES_PHYSICAL)
    ):
        """
        Initialize the PerfScriptParser object.

        Args:
            parsed_script_path (str): The path to the parsed script file.
            threads_num (int, optional): The number of threads to use for parsing. Defaults to min(12, NUM_CORES_PHYSICAL).

        Raises:
            FileNotFoundError: If the parsed script file does not exist.
        """
        if not os.path.exists(parsed_script_path):
            logger.error(f"File not found: {parsed_script_path}")
            raise FileNotFoundError()
        self._perf_script_data = parse_perf_script_file(parsed_script_path, threads_num)

        if len(self._perf_script_data) == 0:
            logger.warning("No data found in the perf script file.")

        if len(self._perf_script_data) >= 10**6:
            logger.info(f"Using pandarallel to speed up, threads_num is {threads_num}.")
            pandarallel.initialize(nb_workers=threads_num)

        self._df_wider = None

    def get_raw_data(self):
        """
        Returns the raw data of the perf script.

        Returns:
            The raw data of the perf script.
        """
        return self._perf_script_data

    def get_wider_data(self):
        """
        Returns a wider version of the performance script data, where the cycles and instructions
        are merged into a single DataFrame. The DataFrame includes columns for command, pid, cpu,
        time, cycles, instructions, addr, symbol, dso_short_name, and CPI (cycles per instruction).

        If the wider data has already been computed, it is returned from the cache. Otherwise, the
        wider data is computed by merging the cycles and instructions DataFrames and calculating
        the CPI.

        Returns:
            pandas.DataFrame: The wider version of the performance script data.
        """
        if self._df_wider is not None:
            return self._df_wider

        df = self._perf_script_data
        df_cycles = df.query("event == 'cycles'")
        df_insns = df.query("event == 'instructions'")
        df_wider = (
            df_cycles.merge(
                df_insns,
                on=[
                    "time",
                    "cpu",
                    "command",
                    "pid",
                    "addr",
                    "symbol",
                    "dso_short_name",
                ],
                suffixes=("_cycles", "_insns"),
            )
            .drop(columns=["event_cycles", "event_insns"])
            .rename(
                {
                    "value_cycles": "cycles",
                    "value_insns": "instructions",
                },
                axis=1,
            )
        )

        df_wider = df_wider[
            [
                "command",
                "pid",
                "cpu",
                "time",
                "cycles",
                "instructions",
                "addr",
                "symbol",
                "dso_short_name",
            ]
        ]
        df_wider["CPI"] = df_wider["cycles"] / df_wider["instructions"]
        self._df_wider = df_wider
        return df_wider

    def get_tidy_data(self, thread_list: list = None):
        """
        Returns a tidy version of the data by pivoting the wider data and renaming the columns.

        Args:
            thread_list (list): A list of hardware thread names to include in the tidy data. If None, all
                threads are included.

        Returns:
            pandas.DataFrame: A tidy version of the data.
        """
        df_wider = self.get_wider_data()

        if thread_list is not None:
            thread_list = [int(thread) for thread in thread_list]
            df_wider = df_wider[df_wider["cpu"].isin(thread_list)]
            if len(thread_list) == 1:
                return df_wider
        df_t = df_wider.pivot_table(
            index=["time"], columns="cpu", aggfunc="first"
        ).reset_index()
        df_t.columns = [f"{col[0]}_{col[1]}" for col in df_t.columns]
        df_t.rename(columns={"time_": "time"}, inplace=True)
        return df_t


def parse_one_line(line):
    """
    Parses a single line of input and returns a list of parsed values.

    Args:
        line (str): The input line to be parsed.

    Returns:
        list: A list containing the parsed values from the input line.
            The list contains the following elements in order:
            - command (str): The command name.
            - pid (int): The process ID.
            - tid (int): The thread ID.
            - cpu (int): The CPU number.
            - time (str): The time value.
            - value (int): The event value.
            - event (str): The event name.
            - addr (str): The address value.
            - symbol (str): The symbol name.
            - caller (str): The caller value.
    """
    try:
        try:
            pattern = r"(\S+|\:-\d+)\s+(\d+|-\d+)\s+\[(\d+)]\s+(\d+\.\d+):\s+(\d+)\s+(\S+):\s+(\S+)\s+(.*?)\s+\((\S+)\)"

            command, pid, cpu, time, value, event, addr, symbol, caller = re.match(
                pattern, line.strip()
            ).groups()
        except Exception as e:
            pattern = r"(\d+|-\d+)\s+\[(\d+)]\s+(\d+\.\d+):\s+(\d+)\s+(\S+):\s+(\S+)\s+(.*?)\s+\((\S+)\)"
            (
                pid,
                cpu,
                time,
                value,
                event,
                addr,
                symbol,
                caller,
            ) = re.match(pattern, line[15:].strip()).groups()

            command = line[:15].strip()

    except Exception as e:
        logger.warning("parse failed for line: " + line + "\n with error: " + str(e))
        return None

    return [
        command,
        int(pid),
        int(cpu),
        time,
        int(value),
        event,
        addr,
        symbol,
        caller,
    ]


def parse_perf_script_file(parsed_script_path: str, processes_num=NUM_CORES_PHYSICAL):
    """
    Parses a perf script file and returns the data as a pandas DataFrame.

    Args:
        parsed_script_path (str): The path to the perf script file.

    Returns:
        pandas.DataFrame: The parsed data as a DataFrame.
    """

    # Open the perf script file in read mode
    with open(parsed_script_path, "r") as file:
        # Read all lines from the file and remove leading/trailing whitespaces
        content = [l.strip() for l in file.readlines()]

    # Ensure that the content is not None
    if content is None:
        logger.info("content is None")
        return None

    # Define the separator used to split the header and table sections
    SEPARATOR = "# ========"

    # Check if the first line is the separator, and remove it if present
    if content[0] == SEPARATOR:
        content.pop(0)

    if SEPARATOR in content:
        # Find the index of the separator in the content list
        sep_idx = content.index(SEPARATOR)

        # Split the content into header and table sections
        header, table = content[:sep_idx], content[sep_idx + 2 :]
    else:
        table = content

    with multiprocessing.Pool(processes=processes_num) as pool:
        data = pool.map(parse_one_line, table)

    data = [d for d in data if d is not None]

    # Parse each line in the table section and create a DataFrame
    return pd.DataFrame(
        data,
        columns=[
            "command",
            "pid",
            "cpu",
            "time",
            "value",
            "event",
            "addr",
            "symbol",
            "dso_short_name",
        ],
    )


if __name__ == "__main__":
    parse_perf_script_file(input("input perf script text data path:"))



import pytest
from unittest.mock import mock_open
import pandas as pd
from pipa.parser.perf_script import parse_perf_script_file, parse_one_line
from pipa.common.logger import logger


@pytest.fixture
def sample_perf_script():
    return """
# ========
#
            perf 3732494 [000] 954453.309835:       2715       cycles:  ffffffff8ae9ceb6 native_write_msr+0x6 ([kernel.kallsyms])
            perf 3732494 [000] 954453.309835:        380 instructions:  ffffffff8ae9ceb6 native_write_msr+0x6 ([kernel.kallsyms])
            perf 3732494 [001] 954453.310054:       2034       cycles:  ffffffff8ae9ceb6 native_write_msr+0x6 ([kernel.kallsyms])
            perf 3732494 [001] 954453.310054:        380 instructions:  ffffffff8ae9ceb6 native_write_msr+0x6 ([kernel.kallsyms])
            perf 3732494 [002] 954453.310232:       1874       cycles:  ffffffff8ae9ceb6 native_write_msr+0x6 ([kernel.kallsyms])
        """


@pytest.fixture
def expected_data():
    return pd.DataFrame(
        {
            "command": ["perf", "perf", "perf", "perf", "perf"],
            "pid": [3732494, 3732494, 3732494, 3732494, 3732494],
            "cpu": [0, 0, 1, 1, 2],
            "time": [
                "954453.309835",
                "954453.309835",
                "954453.310054",
                "954453.310054",
                "954453.310232",
            ],
            "value": [2715, 380, 2034, 380, 1874],
            "event": ["cycles", "instructions", "cycles", "instructions", "cycles"],
            "addr": [
                "ffffffff8ae9ceb6",
                "ffffffff8ae9ceb6",
                "ffffffff8ae9ceb6",
                "ffffffff8ae9ceb6",
                "ffffffff8ae9ceb6",
            ],
            "symbol": [
                "native_write_msr+0x6",
                "native_write_msr+0x6",
                "native_write_msr+0x6",
                "native_write_msr+0x6",
                "native_write_msr+0x6",
            ],
            "dso_short_name": [
                "[kernel.kallsyms]",
                "[kernel.kallsyms]",
                "[kernel.kallsyms]",
                "[kernel.kallsyms]",
                "[kernel.kallsyms]",
            ],
        }
    )


def test_parse_perf_script_file(monkeypatch, sample_perf_script, expected_data):
    monkeypatch.setattr("builtins.open", mock_open(read_data=""))
    monkeypatch.setattr("os.path.exists", lambda x: True)
    open_mock = mock_open(read_data=sample_perf_script)
    monkeypatch.setattr("builtins.open", open_mock)

    # Call the function with the mocked data
    result_df = parse_perf_script_file("dummy_path")

    # Check if the resulting DataFrame matches the expected data
    pd.testing.assert_frame_equal(result_df, expected_data)


@pytest.fixture
def test_line():
    return "            perf 3732494 [000] 954453.309835:       2715       cycles:  ffffffff8ae9ceb6 native_write_msr+0x6 ([kernel.kallsyms])"


@pytest.fixture
def expected_output():
    return [
        "perf",
        3732494,
        0,
        "954453.309835",
        2715,
        "cycles",
        "ffffffff8ae9ceb6",
        "native_write_msr+0x6",
        "[kernel.kallsyms]",
    ]


def test_parse_one_line(monkeypatch, test_line, expected_output):
    monkeypatch.setattr(logger, "warning", lambda msg: None)  # Mocking logger.warning
    result = parse_one_line(test_line)
    assert result == expected_output


def test_parse_one_line_invalid_format(monkeypatch):
    invalid_line = "invalid format line"
    result = parse_one_line(invalid_line)
    assert result is None


if __name__ == "__main__":
    pytest.main([__file__])



import questionary
from pipa.common.hardware.cpu import get_cpu_cores
from rich import print
from io import TextIOWrapper
from datetime import datetime


import os
import yaml


def ask_number(question: str, default: int) -> int:
    """
    Asks the user to input a number based on the given question and default value.

    Args:
        question (str): The question to ask the user.
        default (int): The default value to return if the user doesn't input anything.

    Returns:
        int: The number inputted by the user or the default value.
    """
    result = questionary.text(question, str(default)).ask().strip()

    if result == "":
        return default
    elif result.isdigit():
        return int(result)
    else:
        print("Please input a number.")
        exit(1)


CORES_ALL = get_cpu_cores()


def quest_basic():
    """
    Asks the user a series of questions to gather basic information for data storage and performance recording.

    Returns:
        dict: A dictionary containing the user's choices for workspace, frequency of perf-record, event of perf-record,
              whether to use perf-annotate, whether to use perf-stat or emon, count deltas of perf-stat, events of perf-stat,
              and the location of mpp if emon is chosen.

    Raises:
        ValueError: If an invalid choice is made.
    """
    workspace = questionary.text(
        "Where do you want to store your data? (Default: ./)\n", "./"
    ).ask()

    if workspace == "":
        workspace = "./"

    if not os.path.exists(workspace):
        os.makedirs(workspace)

    freq_record = ask_number(
        "What's the frequency of perf-record? (Default: 999)\n", 999
    )
    events_record = questionary.text(
        "What's the event of perf-record? (Default: {cycles,instructions}:S)\n",
        "{cycles,instructions}:S",
    ).ask()

    annotete = questionary.select(
        "Whether to use perf-annotate?\n", choices=["Yes", "No"], default="No"
    ).ask()
    annotete = True if annotete == "Yes" else False

    stat = questionary.select(
        "Do you want to use perf-stat or emon?\n",
        choices=["perf-stat", "emon"],
        default="perf-stat",
    ).ask()

    if stat == "perf-stat":
        freq_stat = ask_number(
            "What's count deltas of perf-stat? (Default: 1000 milliseconds)\n", 1000
        )
        events_stat = questionary.text(
            "What's the event of perf-stat?\n",
            "cycles,instructions,branch-misses,L1-dcache-load-misses,L1-icache-load-misses",
        ).ask()
        return {
            "workspace": workspace,
            "freq_record": freq_record,
            "events_record": events_record,
            "use_emon": False,
            "count_delta_stat": freq_stat,
            "events_stat": events_stat,
            "annotete": annotete,
        }
    elif stat == "emon":
        mpp = questionary.text(
            "Where is the mpp?\n",
            "/mnt/hdd/share/emon/system_health_monitor",
        ).ask()
        return {
            "workspace": workspace,
            "freq_record": freq_record,
            "events_record": events_record,
            "use_emon": True,
            "mpp": mpp,
            "annotete": annotete,
        }

    raise ValueError("Invalid choice.")


def write_title(file: TextIOWrapper):
    """
    Writes the title section to the given file.

    Args:
        file (TextIOWrapper): The file object to write the title section to.
    """
    current_time = datetime.now().isoformat()
    file.write(
        f"""#!/bin/bash
# The script generated by PIPA-TREE is used to collect performance data.
# Please check whether it meets expectations before running.
# ZJU SPAIL(https://github.com/ZJU-SPAIL) reserves all rights.
# Generated at {current_time}

# Check if sar and perf are available
if ! command -v sar &> /dev/null; then
echo "sar command not found. Please install sar."
exit 1
fi

if ! command -v perf &> /dev/null; then
echo "perf command not found. Please install perf."
exit 1
fi\n\n"""
    )


def load_yaml_config(file_path: str = "config-pipa-shu.yaml") -> dict:
    """
    Parses a YAML file and returns the contents as a dictionary.

    Args:
        file_path (str): The path to the YAML file.

    Returns:
        dict: The contents of the YAML file as a dictionary.
    """
    with open(file_path, "r") as file:
        data = yaml.safe_load(file)
    return data


def opener(path, flags):
    """
    Opens a file at the given path with the specified flags.
    Mode is set to 0o755.

    Args:
        path (str): The path to the file.
        flags (int): The flags to use when opening the file.

    Returns:
        int: The file descriptor of the opened file.
    """
    descriptor = os.open(path=path, flags=flags, mode=0o755)
    return descriptor



import pytest
from unittest.mock import mock_open
import pandas as pd
from pipa.parser.perf_report import parse_perf_report_file, parse_one_line


@pytest.fixture
def sample_data():
    return """
#         Overhead  Command          Shared Object                                                       Symbol
# ................  ...............  ..................................................................  ..................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................................
#
    14.31%   2.91%  java             libjvm.so                                                           [.] SpinPause
     7.10%   3.77%  swapper          [kernel.kallsyms]                                                   [k] mwait_idle_with_hints.constprop.0
     2.08%   0.43%  java             libjvm.so                                                           [.] ParallelTaskTerminator::offer_termination
     1.78%   2.71%  perf             perf                                                                [.] append_chain_children
     1.43%   3.06%  node             watcher.node                                                        [.] std::__detail::_Executor<__gnu_cxx::__normal_iterator<char const*, std::__cxx11::basic_string<char, std::char_traits<char>, std::allocator<char> > >, std::allocator<std::__cxx11::sub_match<__gnu_cxx::__normal_iterator<char const*, std::__cxx11::basic_string<char, std::char_traits<char>, std::allocator<char> > > > >, std::cxx11::regex_traits<char>, true>::_M_dfs
     1.09%   2.34%  node             watcher.node                                                        [.] std::__detail::_Executor<__gnu_cxx::__normal_iterator<char const*, std::__cxx11::basic_string<char, std::char_traits<char>, std::allocator<char> > >, std::allocator<std::__cxx11::sub_match<__gnu_cxx::__normal_iterator<char const*, std::__cxx11::basic_string<char, std::char_traits<char>, std::allocator<char> > > > >, std::__cxx11::regex_traits<char>, true>::_M_lookahead
     0.95%   0.41%  java             libjvm.so                                                           [.] GenericTaskQueueSet<OverflowTaskQueue<StarTask, (MemoryType)5, 131072u>, (MemoryType)5>::steal_best_of_2
     0.79%   1.18%  htop             [kernel.kallsyms]                                                   [k] __d_lookup_rcu
    """


@pytest.fixture
def expected_df():
    data = {
        "overhead_cycles": [14.31, 7.10, 2.08, 1.78, 1.43, 1.09, 0.95, 0.79],
        "overhead_insns": [2.91, 3.77, 0.43, 2.71, 3.06, 2.34, 0.41, 1.18],
        "command": [
            "java",
            "swapper",
            "java",
            "perf",
            "node",
            "node",
            "java",
            "htop",
        ],
        "shared_object": [
            "libjvm.so",
            "[kernel.kallsyms]",
            "libjvm.so",
            "perf",
            "watcher.node",
            "watcher.node",
            "libjvm.so",
            "[kernel.kallsyms]",
        ],
        "execution_mode": [".", "k", ".", ".", ".", ".", ".", "k"],
        "symbol": [
            "SpinPause",
            "mwait_idle_with_hints.constprop.0",
            "ParallelTaskTerminator::offer_termination",
            "append_chain_children",
            "std::__detail::_Executor<__gnu_cxx::__normal_iterator<char const*, std::__cxx11::basic_string<char, std::char_traits<char>, std::allocator<char> > >, std::allocator<std::__cxx11::sub_match<__gnu_cxx::__normal_iterator<char const*, std::__cxx11::basic_string<char, std::char_traits<char>, std::allocator<char> > > > >, std::cxx11::regex_traits<char>, true>::_M_dfs",
            "std::__detail::_Executor<__gnu_cxx::__normal_iterator<char const*, std::__cxx11::basic_string<char, std::char_traits<char>, std::allocator<char> > >, std::allocator<std::__cxx11::sub_match<__gnu_cxx::__normal_iterator<char const*, std::__cxx11::basic_string<char, std::char_traits<char>, std::allocator<char> > > > >, std::__cxx11::regex_traits<char>, true>::_M_lookahead",
            "GenericTaskQueueSet<OverflowTaskQueue<StarTask, (MemoryType)5, 131072u>, (MemoryType)5>::steal_best_of_2",
            "__d_lookup_rcu",
        ],
    }
    return pd.DataFrame(data)


def test_parse_perf_report_file(monkeypatch, sample_data, expected_df):
    # Mock file read
    mock_open_func = mock_open(read_data=sample_data)
    monkeypatch.setattr("builtins.open", mock_open_func)

    # Call the function
    result = parse_perf_report_file("./test_resources/perf.report")

    for col in result.columns:
        for i, (act, exp) in enumerate(zip(result[col], expected_df[col])):
            if act != exp:
                print(f"Column: {col}, Row: {i}, Actual: {act}, Expected: {exp}")

    # Compare the result with the expected DataFrame
    pd.testing.assert_frame_equal(result, expected_df)


@pytest.fixture
def test_line():
    return "    14.31%   2.91%  java             libjvm.so                                                           [.] SpinPause"


@pytest.fixture
def line_ranges():
    return [(4, 18), (20, 24), (37, 46), (105, -1)]


def test_parse_one_line(test_line, line_ranges):
    expected_output = (14.31, 2.91, "java", "libjvm.so", ".", "SpinPause")
    result = parse_one_line(test_line, line_ranges)
    assert result == expected_output


def test_parse_one_line_incorrect_format():
    incorrect_line = "incorrect format data"
    line_ranges = [(4, 18), (20, 24), (37, 46), (105, -1)]
    result = parse_one_line(incorrect_line, line_ranges)
    assert result is None


if __name__ == "__main__":
    pytest.main([__file__])



import questionary
import os
from rich import print

CONFIG_TEMPLATE = """# PIPA-Shu Configuration
# Use pipa generate to generate data collection script based on this configuration.
# Command Example: pipa generate --config_path=./data/config-pipa-shu.yaml

# PIPA (Platform Integrated Performance Analytics)
# Developed by: SPAIL, ZJU https://github.com/ZJU-SPAIL
# All rights reserved.

workspace: ./data # The workspace to store the performance data and collecting script. Will be created if not exists.
freq_record: 999 # The frequency of perf-record.
events_record: "{cycles,instructions}:S" # The events to be used in perf-record.
annotete: False # Whether to annotate the performance data.

use_emon: False # Whether to use emon.

count_delta_stat: 1000 # The count delta of perf-stat.
# No need to set if use_emon is True.
events_stat: # The events to be used in perf-stat.
        - cycles
        - instructions
        - branch-misses
        - L1-dcache-load-misses
        - L1-icache-load-misses
# No need to set if use_emon is True.

MPP_HOME: /mnt/hdd/share/emon/system_health_monitor # The path to the mpp.
# No need to set if use_emon is False.

run_by_perf: False # Whether to run the workload by perf.

duration_record: 120 # The duration of the perf-record.
# No need to set if run_by_perf is True.
# Set it -1 if you want to run the perf until workload finishes or you kill perf manually.
duration_stat: 120 # The duration of the perf-stat.
# No need to set if run_by_perf is True.
# Set it -1 if you want to run the perf until workload finishes or you kill perf manually.

use_taskset: True # Whether to use taskset to bind the process to specific cores.
# No need to set if run_by_perf is False.
core_range: 0-7 # The range of cores to be used.
# No need to set if use_taskset or run_by_perf is False.
command: perf bench futex hash # The command to run workload.
# No need to set if run_by_perf is False.

export_config: True # Whether to export the configuration after parsing.

"""


def query_filepath():
    """
    Prompt the user to enter the filepath where they want to store the configuration template.

    Returns:
        str: The filepath entered by the user.
    """
    workspace = questionary.text(
        "Where do you want to store the configuration template? (Default: ./)\n", "./"
    ).ask()
    if workspace == "":
        workspace = "./"
    return workspace


def generate_template(content: str, filename: str):
    """
    Generate a template file with the given content and filename.

    Args:
        content (str): The content to be written to the template file.
        filename (str): The name of the template file.

    Returns:
        int: The number of characters written to the template file.
    """
    workspace = query_filepath()
    config_file_path = os.path.join(workspace, filename)
    if not os.path.exists(workspace):
        os.makedirs(workspace)
    with open(config_file_path, "w") as f:
        result = f.write(content)
    print(f"Template has been generated at {config_file_path}")
    return result


def generate_pipashu_template():
    """
    Generates a Pipa Shu configuration template.

    Returns:
        int: The number of characters written to the template file.
    """
    return generate_template(CONFIG_TEMPLATE, "config-pipa-shu.yaml")


UPLOAD_TEMPLATE = """# PIPA-Shu Upload Configuration
# Use pipa upload to upload the data to PIPAD server based on this configuration.
# Command Example: pipa upload --config_path=./data/config-upload.yaml
workload: rocksdb
# The name of the workload.
transaction: 7561946
# The number of transactions.
data_location: /path/to/data/collected/by/pipashu
# The location of the data collected by PIPAShu.

cores: [36, 37, 38, 39]
# The numbers of logical cores used in the workload.
dev: sdc
# The used disk device name.

hw_info: 1*4*1
# The hardware configuration (sockets*cores*SMT).
sw_info: RocksDB 7.9.2 build in release mode, debug_level=0, threads_num=16, db_bench with benchmark.sh
# The software configuration.

platform: Intel SPR 4510
# The platform user used.
cpu_frequency_mhz: 2600
# The CPU frequency in MHz.
# Only needed when the platform is Huawei.

comment: "This is a template for the upload configuration."
# Any comments.
pipad_addr: 10.82.77.113
# The PIPAD server address.
pipad_port: 50051
# The PIPAD server port.
"""


def generate_upload_template(filename: str = "config-upload.yaml"):
    """
    Generate an upload template for PipaSHU configuration.

    Args:
        filename (str, optional): The name of the output file. Defaults to "config-upload.yaml".

    Returns:
        str: The generated template as a string.
    """
    return generate_template(UPLOAD_TEMPLATE, filename)



import logging

# Create a logger
logger = logging.getLogger(__name__)

# Set the logging level
logger.setLevel(logging.INFO)

# Create a stream handler and set its level
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.WARNING)

# Create a formatter and add it to the stream handler
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
stream_handler.setFormatter(formatter)

# Add the stream handler to the logger
logger.addHandler(stream_handler)

if __name__ == "__main__":
    # Example usage
    logger.debug("This is a debug message")
    logger.info("This is an info message")
    logger.warning("This is a warning message")
    logger.error("This is an error message")
    logger.critical("This is a critical message")



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class DeployRequest(_message.Message):
    __slots__ = (
        "workload",
        "transactions",
        "throughput",
        "used_threads",
        "run_time",
        "cycles",
        "instructions",
        "cycles_per_second",
        "instructions_per_second",
        "CPI",
        "cycles_per_requests",
        "path_length",
        "cpu_frequency_mhz",
        "cpu_usr",
        "cpu_nice",
        "cpu_sys",
        "cpu_iowait",
        "cpu_steal",
        "cpu_irq",
        "cpu_soft",
        "cpu_guest",
        "cpu_gnice",
        "cpu_idle",
        "cpu_util",
        "kbmemfree",
        "kbavail",
        "kbmemused",
        "percent_memused",
        "kbbuffers",
        "kbcached",
        "kbcommit",
        "percent_commit",
        "kbactive",
        "kbinact",
        "kbdirty",
        "kbanonpg",
        "kbslab",
        "kbkstack",
        "kbpgtbl",
        "kbvmused",
        "dev",
        "tps",
        "rkB_s",
        "wkB_s",
        "dkB_s",
        "areq_sz",
        "aqu_sz",
        "disk_await",
        "percent_disk_util",
        "data_location",
        "hw_info",
        "sw_info",
        "platform",
        "comment",
        "username",
    )
    WORKLOAD_FIELD_NUMBER: _ClassVar[int]
    TRANSACTIONS_FIELD_NUMBER: _ClassVar[int]
    THROUGHPUT_FIELD_NUMBER: _ClassVar[int]
    USED_THREADS_FIELD_NUMBER: _ClassVar[int]
    RUN_TIME_FIELD_NUMBER: _ClassVar[int]
    CYCLES_FIELD_NUMBER: _ClassVar[int]
    INSTRUCTIONS_FIELD_NUMBER: _ClassVar[int]
    CYCLES_PER_SECOND_FIELD_NUMBER: _ClassVar[int]
    INSTRUCTIONS_PER_SECOND_FIELD_NUMBER: _ClassVar[int]
    CPI_FIELD_NUMBER: _ClassVar[int]
    CYCLES_PER_REQUESTS_FIELD_NUMBER: _ClassVar[int]
    PATH_LENGTH_FIELD_NUMBER: _ClassVar[int]
    CPU_FREQUENCY_MHZ_FIELD_NUMBER: _ClassVar[int]
    CPU_USR_FIELD_NUMBER: _ClassVar[int]
    CPU_NICE_FIELD_NUMBER: _ClassVar[int]
    CPU_SYS_FIELD_NUMBER: _ClassVar[int]
    CPU_IOWAIT_FIELD_NUMBER: _ClassVar[int]
    CPU_STEAL_FIELD_NUMBER: _ClassVar[int]
    CPU_IRQ_FIELD_NUMBER: _ClassVar[int]
    CPU_SOFT_FIELD_NUMBER: _ClassVar[int]
    CPU_GUEST_FIELD_NUMBER: _ClassVar[int]
    CPU_GNICE_FIELD_NUMBER: _ClassVar[int]
    CPU_IDLE_FIELD_NUMBER: _ClassVar[int]
    CPU_UTIL_FIELD_NUMBER: _ClassVar[int]
    KBMEMFREE_FIELD_NUMBER: _ClassVar[int]
    KBAVAIL_FIELD_NUMBER: _ClassVar[int]
    KBMEMUSED_FIELD_NUMBER: _ClassVar[int]
    PERCENT_MEMUSED_FIELD_NUMBER: _ClassVar[int]
    KBBUFFERS_FIELD_NUMBER: _ClassVar[int]
    KBCACHED_FIELD_NUMBER: _ClassVar[int]
    KBCOMMIT_FIELD_NUMBER: _ClassVar[int]
    PERCENT_COMMIT_FIELD_NUMBER: _ClassVar[int]
    KBACTIVE_FIELD_NUMBER: _ClassVar[int]
    KBINACT_FIELD_NUMBER: _ClassVar[int]
    KBDIRTY_FIELD_NUMBER: _ClassVar[int]
    KBANONPG_FIELD_NUMBER: _ClassVar[int]
    KBSLAB_FIELD_NUMBER: _ClassVar[int]
    KBKSTACK_FIELD_NUMBER: _ClassVar[int]
    KBPGTBL_FIELD_NUMBER: _ClassVar[int]
    KBVMUSED_FIELD_NUMBER: _ClassVar[int]
    DEV_FIELD_NUMBER: _ClassVar[int]
    TPS_FIELD_NUMBER: _ClassVar[int]
    RKB_S_FIELD_NUMBER: _ClassVar[int]
    WKB_S_FIELD_NUMBER: _ClassVar[int]
    DKB_S_FIELD_NUMBER: _ClassVar[int]
    AREQ_SZ_FIELD_NUMBER: _ClassVar[int]
    AQU_SZ_FIELD_NUMBER: _ClassVar[int]
    DISK_AWAIT_FIELD_NUMBER: _ClassVar[int]
    PERCENT_DISK_UTIL_FIELD_NUMBER: _ClassVar[int]
    DATA_LOCATION_FIELD_NUMBER: _ClassVar[int]
    HW_INFO_FIELD_NUMBER: _ClassVar[int]
    SW_INFO_FIELD_NUMBER: _ClassVar[int]
    PLATFORM_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    workload: str
    transactions: int
    throughput: float
    used_threads: _containers.RepeatedScalarFieldContainer[int]
    run_time: float
    cycles: int
    instructions: int
    cycles_per_second: float
    instructions_per_second: float
    CPI: float
    cycles_per_requests: float
    path_length: float
    cpu_frequency_mhz: float
    cpu_usr: float
    cpu_nice: float
    cpu_sys: float
    cpu_iowait: float
    cpu_steal: float
    cpu_irq: float
    cpu_soft: float
    cpu_guest: float
    cpu_gnice: float
    cpu_idle: float
    cpu_util: float
    kbmemfree: int
    kbavail: int
    kbmemused: int
    percent_memused: float
    kbbuffers: int
    kbcached: int
    kbcommit: int
    percent_commit: float
    kbactive: int
    kbinact: int
    kbdirty: int
    kbanonpg: int
    kbslab: int
    kbkstack: int
    kbpgtbl: int
    kbvmused: int
    dev: str
    tps: float
    rkB_s: float
    wkB_s: float
    dkB_s: float
    areq_sz: float
    aqu_sz: float
    disk_await: float
    percent_disk_util: float
    data_location: str
    hw_info: str
    sw_info: str
    platform: str
    comment: str
    username: str
    def __init__(
        self,
        workload: _Optional[str] = ...,
        transactions: _Optional[int] = ...,
        throughput: _Optional[float] = ...,
        used_threads: _Optional[_Iterable[int]] = ...,
        run_time: _Optional[float] = ...,
        cycles: _Optional[int] = ...,
        instructions: _Optional[int] = ...,
        cycles_per_second: _Optional[float] = ...,
        instructions_per_second: _Optional[float] = ...,
        CPI: _Optional[float] = ...,
        cycles_per_requests: _Optional[float] = ...,
        path_length: _Optional[float] = ...,
        cpu_frequency_mhz: _Optional[float] = ...,
        cpu_usr: _Optional[float] = ...,
        cpu_nice: _Optional[float] = ...,
        cpu_sys: _Optional[float] = ...,
        cpu_iowait: _Optional[float] = ...,
        cpu_steal: _Optional[float] = ...,
        cpu_irq: _Optional[float] = ...,
        cpu_soft: _Optional[float] = ...,
        cpu_guest: _Optional[float] = ...,
        cpu_gnice: _Optional[float] = ...,
        cpu_idle: _Optional[float] = ...,
        cpu_util: _Optional[float] = ...,
        kbmemfree: _Optional[int] = ...,
        kbavail: _Optional[int] = ...,
        kbmemused: _Optional[int] = ...,
        percent_memused: _Optional[float] = ...,
        kbbuffers: _Optional[int] = ...,
        kbcached: _Optional[int] = ...,
        kbcommit: _Optional[int] = ...,
        percent_commit: _Optional[float] = ...,
        kbactive: _Optional[int] = ...,
        kbinact: _Optional[int] = ...,
        kbdirty: _Optional[int] = ...,
        kbanonpg: _Optional[int] = ...,
        kbslab: _Optional[int] = ...,
        kbkstack: _Optional[int] = ...,
        kbpgtbl: _Optional[int] = ...,
        kbvmused: _Optional[int] = ...,
        dev: _Optional[str] = ...,
        tps: _Optional[float] = ...,
        rkB_s: _Optional[float] = ...,
        wkB_s: _Optional[float] = ...,
        dkB_s: _Optional[float] = ...,
        areq_sz: _Optional[float] = ...,
        aqu_sz: _Optional[float] = ...,
        disk_await: _Optional[float] = ...,
        percent_disk_util: _Optional[float] = ...,
        data_location: _Optional[str] = ...,
        hw_info: _Optional[str] = ...,
        sw_info: _Optional[str] = ...,
        platform: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        username: _Optional[str] = ...,
    ) -> None: ...

class DeployResp(_message.Message):
    __slots__ = (
        "message",
        "username",
        "time",
        "hash",
        "upload_datetime",
        "status_code",
    )
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    TIME_FIELD_NUMBER: _ClassVar[int]
    HASH_FIELD_NUMBER: _ClassVar[int]
    UPLOAD_DATETIME_FIELD_NUMBER: _ClassVar[int]
    STATUS_CODE_FIELD_NUMBER: _ClassVar[int]
    message: str
    username: str
    time: str
    hash: str
    upload_datetime: str
    status_code: int
    def __init__(
        self,
        message: _Optional[str] = ...,
        username: _Optional[str] = ...,
        time: _Optional[str] = ...,
        hash: _Optional[str] = ...,
        upload_datetime: _Optional[str] = ...,
        status_code: _Optional[int] = ...,
    ) -> None: ...



import os
from pipa.common.config import ALL_PATH
from pipa.common.logger import logger


def create_directories():
    # Create the directories
    for path in ALL_PATH:
        os.makedirs(path, exist_ok=True)
        logger.info(f"Directory created: {path}")
    logger.info("All directories created successfully.")


if __name__ == "__main__":
    create_directories()



import questionary
from rich import print
import os
from pipa.service.pipashu import PIPAShuData
from pipa.common.logger import logger
from pipa.service.pipad.pipad_client import PIPADClient
import pipa.service.pipad.pipad_pb2 as pipadlib
import getpass
import yaml


def check_workload(workload):
    """
    Check if the workload name is valid.

    Args:
        workload (str): The name of the workload.

    Returns:
        None

    Raises:
        SystemExit: If the workload name is empty.

    """
    if workload == "":
        print("Please input a valid workload name.")
        exit(1)


def check_transaction(transaction):
    """
    Check if the given transaction is a valid number.

    Args:
        transaction (str): The transaction to be checked.

    Returns:
        int: The transaction converted to an integer if it is a valid number.

    Raises:
        ValueError: If the transaction is not a valid number.

    """
    if not transaction.isdigit():
        print("Please input a valid number.")
        exit(1)
    try:
        transaction = int(transaction)
    except ValueError:
        print("Please input a valid number.")
        raise ValueError
    return transaction


def check_path(data_location):
    """
    Check if the specified data location is valid.

    Args:
        data_location (str): The path to the data location.

    Raises:
        SystemExit: If the data location is empty or does not exist.

    """
    if data_location == "":
        print("Please input a valid data location.")
        exit(1)
    elif not os.path.exists(data_location):
        print("The folder does not exist.")
        exit(1)


def check_cores(cores):
    """
    Check if the given cores are valid.

    Args:
        cores (str): A string representing the core numbers separated by commas.

    Returns:
        list: A list of integers representing the valid core numbers.

    Raises:
        ValueError: If the input is not a valid core number.

    """
    if cores == "":
        print("Please input a valid core number.")
        exit(1)
    try:
        cores = cores.split(",")
        cores = list(map(int, cores))
    except ValueError:
        print("Please input a valid core number.")
        raise ValueError
    except Exception as e:
        raise e

    return cores


def quest():
    """
    Collects information from the user and returns a dictionary containing the collected data.

    Returns:
        dict: A dictionary containing the collected information with the following keys:
            - 'workload': The name of the workload.
            - 'transaction': The number of transactions.
            - 'data_location': The location of the data collected by PIPAShu.
            - 'cores': The threads used in the workload.
            - 'dev': The used disk device name.
            - 'hw_info': The hardware configuration (sockets*cores*SMT).
            - 'sw_info': The software configuration.
            - 'platform': The platform.
            - 'comment': Any comments.
            - 'pipad_addr': The PIPAD server address.
            - 'pipad_port': The PIPAD server port.
    """
    workload = questionary.text("What's the name of workload?").ask()
    check_workload(workload)

    transaction = questionary.text("What's the number of transaction?").ask()
    transaction = check_transaction(transaction)

    data_location = questionary.text("Where's the data collected by PIPAShu?").ask()
    check_path(data_location)

    cores = questionary.text(
        "What are the threads used in the workload? Split by ,", default="0,1,2,3"
    ).ask()
    cores = check_cores(cores)

    dev = questionary.text("What's the used disk device name?").ask()
    if dev == "":
        dev = None

    hw_info = questionary.text(
        "What's the hardware configuration (sockets*cores*SMT)?", default="1*1*1"
    ).ask()

    sw_info = questionary.text("What's the software configuration?").ask()

    platform = questionary.text("What's the platform?", default="IceLake 8383C").ask()

    cpu_frequency_mhz = None
    if "huawei" in platform.lower():
        cpu_frequency_mhz = questionary.text(
            "What's the CPU frequency in MHz?", default="2600"
        ).ask()
        cpu_frequency_mhz = int(cpu_frequency_mhz)

    comment = questionary.text("Any comments?").ask()

    pipad_server = questionary.text("What's the PIPAD server address?").ask()

    pipad_port = questionary.text("What's the PIPAD server port?").ask()

    return {
        "workload": workload,
        "transaction": transaction,
        "data_location": data_location,
        "cores": cores,
        "dev": dev,
        "hw_info": hw_info,
        "sw_info": sw_info,
        "platform": platform,
        "cpu_frequency_mhz": cpu_frequency_mhz,
        "comment": comment,
        "pipad_addr": pipad_server,
        "pipad_port": pipad_port,
    }


def build_with_config_path(config_path: str):
    """
    Build with a configuration file path.

    Args:
        config_path (str): The path to the configuration file.

    Returns:
        The result of the build process.
    """
    with open(config_path, "r") as f:
        config = yaml.safe_load(f)
    return build(config)


def build(config: dict):
    """
    Builds the data for upload based on the given configuration.

    Args:
        config (dict): A dictionary containing the configuration parameters.

    Returns:
        dict: A dictionary containing the built data for upload.
            - workload: The name of the workload.
            - transaction: The number of transactions.
            - data_location: The location of the data collected by PIPAShu.
            - cores: The hardware cores used in the workload.
            - dev: The used disk device name.
            - hw_info: The hardware configuration.
            - sw_info: The software configuration.
            - platform: The platform used.
            - comment: Any comments.
            - pipad_addr: The PIPAD server address. Optional.
            - pipad_port: The PIPAD server port. Optional.
            - perf_stat_path: The path to the perf-stat.csv file. Only required if data_location is not provided.
            - sar_path: The path to the sar.txt file. Only required if data_location is not provided.
            - perf_script_path: The path to the perf.script file. Only required if data_location is not provided.


    Raises:
        Exception: If the required files are missing.

    """
    data_dir = config.get("data_location", None)
    if data_dir:
        perf_stat_path = os.path.join(data_dir, "perf-stat.csv")
        sar_path = os.path.join(data_dir, "sar.txt")
        perf_script_path = os.path.join(data_dir, "perf.script")
    else:
        perf_stat_path = config.get("perf_stat_path", None)
        sar_path = config.get("sar_path", None)
        perf_script_path = config.get("perf_script_path", None)

        config["data_location"] = perf_stat_path.rsplit("/", 1)[0]

    if not os.path.exists(perf_stat_path):
        logger.error("perf-stat.csv does not exist.")
        raise Exception("perf-stat.csv does not exist.")
    elif not os.path.exists(sar_path):
        logger.error("sar.txt does not exist.")
        raise Exception("sar.txt does not exist.")
    elif not perf_script_path and not os.path.exists(perf_script_path):
        logger.warning("perf.script does not exist.")
        perf_script_path = None

    cpu_frequency_mhz = config.get("cpu_frequency_mhz", None)

    data = PIPAShuData(perf_stat_path, sar_path, perf_script_path).get_metrics(
        config["transaction"],
        config["cores"],
        dev=config.get("dev", None),
        freq_MHz=cpu_frequency_mhz,
    )

    config.pop("transaction")
    config.pop("cores")
    if "dev" in config:
        config.pop("dev")

    config["username"] = getpass.getuser()

    result = {**data, **config}
    logger.info("Data built successfully.")
    logger.info(str(result))
    return result


def send(data: dict, addr: str = None, port: int = 50051):
    """
    Sends the provided data to the specified address and port using the PIPADClient.

    Args:
        data (dict): The data to be sent.
        addr (str, optional): The address to send the data to. If not provided, the address will be retrieved from the data dictionary.
        port (int, optional): The port to send the data to. If not provided, the default port 50051 will be used.

    Returns:
        The result of the PIPADClient's deploy method.

    """
    if not addr:
        addr = data["pipad_addr"]
    if not port:
        port = data["pipad_port"]
    req = pipadlib.DeployRequest(
        transactions=data["transactions"],
        throughput=data["throughput"],
        used_threads=data["used_threads"],
        run_time=data["run_time"],
        cycles=data["cycles"],
        instructions=data["instructions"],
        cycles_per_second=data["cycles_per_second"],
        instructions_per_second=data["instructions_per_second"],
        CPI=data["CPI"],
        cycles_per_requests=data["cycles_per_requests"],
        path_length=data["path_length"],
        cpu_frequency_mhz=data["cpu_frequency_mhz"],
        cpu_usr=data[r"%usr"],
        cpu_nice=data[r"%nice"],
        cpu_sys=data[r"%sys"],
        cpu_iowait=data[r"%iowait"],
        cpu_steal=data[r"%steal"],
        cpu_irq=data[r"%irq"],
        cpu_soft=data[r"%soft"],
        cpu_guest=data[r"%guest"],
        cpu_gnice=data[r"%gnice"],
        cpu_idle=data[r"%idle"],
        cpu_util=data[r"%util"],
        kbmemfree=data["kbmemfree"],
        kbavail=data["kbavail"],
        kbmemused=data["kbmemused"],
        percent_memused=data[r"%memused"],
        kbbuffers=data["kbbuffers"],
        kbcached=data["kbcached"],
        kbcommit=data["kbcommit"],
        percent_commit=data[r"%commit"],
        kbactive=data["kbactive"],
        kbinact=data["kbinact"],
        kbdirty=data["kbdirty"],
        kbanonpg=data["kbanonpg"],
        kbslab=data["kbslab"],
        kbkstack=data["kbkstack"],
        kbpgtbl=data["kbpgtbl"],
        kbvmused=data["kbvmused"],
        tps=data["tps"],
        rkB_s=data[r"rkB/s"],
        wkB_s=data[r"wkB/s"],
        dkB_s=data[r"dkB/s"],
        areq_sz=data["areq-sz"],
        aqu_sz=data["aqu-sz"],
        disk_await=data["disk_await"],
        percent_disk_util=data[r"%disk_util"],
        workload=data["workload"],
        data_location=data["data_location"],
        dev=data["DEV"],
        hw_info=data["hw_info"],
        sw_info=data["sw_info"],
        platform=data["platform"],
        comment=data["comment"],
        username=data["username"],
    )

    logger.info(f"Sending data to {addr}:{port}")

    resp = PIPADClient(port, addr).deploy(req)

    if resp is not None:
        if resp.status_code == 200:
            logger.info("Upload success.")
            logger.info(
                f"Message: {resp.message}, Username: {resp.username}, Hash: {resp.hash}, Time: {resp.time}"
            )
        else:
            logger.warning(
                f"Upload failed: {resp.message} Status: {resp.status_code} time: {resp.time}"
            )
    return resp


def load(config_path: str = None, verbose: bool = False):
    """
    Load the configuration from a YAML file or prompt the user for configuration.

    Args:
        config_path (str, optional): Path to the YAML configuration file. Defaults to None.
        verbose (bool, optional): Flag indicating whether to enable verbose mode. Defaults to False.

    Returns:
        dict: The loaded configuration.

    """
    if config_path:
        with open(config_path, "r") as f:
            config = yaml.safe_load(f)
    else:
        config = quest()

    if verbose:
        print("Verbose mode enabled.")
        print("Configuration:", str(config))
    return config


def main(config_path: str = None, verbose: bool = False):
    """
    This is the main function for the upload service in the pipa project.

    Args:
        config_path (str, optional): The path to the configuration file. If not provided, the user will be prompted to enter the configuration.
        verbose (bool, optional): If True, the function will print additional information.

    Returns:
        None
    """

    data = build(load(config_path, verbose))

    if verbose:
        print("Data:", str(data))

    return send(data)


if __name__ == "__main__":
    main()



import pandas as pd
from pipa.common.logger import logger


def parse_one_line(s, lr):
    """
    Parse a single line of a performance report.

    Args:
        s (str): The input line to parse.
        lr (list): A list of tuples representing the start and end indices of each field in the line.

    Returns:
        tuple: A tuple containing the parsed values from the line in the following order:
            - overhead_cycles (float): The number of overhead cycles.
            - overhead_insns (float): The number of overhead instructions.
            - command (str): The command associated with the line.
            - shared_object (str): The shared object associated with the line.
            - execution_mode (str): The execution mode associated with the line.
            - symbol (str): The symbol associated with the line.
    """
    try:
        # TODO support more than two columns for the overhead percentage
        (
            overhead,
            command,
            shared_object,
            symbol,
        ) = (
            s[lr[0][0] : lr[0][1]],
            s[lr[1][0] : lr[1][1]],
            s[lr[2][0] : lr[2][1]],
            s[lr[3][0] :],
        )
        overhead_cycles, overhead_insns = overhead.split()
        execution_mode = symbol.split()[0]
        symbol = " ".join(symbol.split()[1:])
        execution_mode = execution_mode[1]
    except Exception as e:
        logger.warning("parse failed for line: " + s + "\n with error: " + str(e))
        return None
    return (
        float(overhead_cycles[:-1]),
        float(overhead_insns[:-1]),
        command.strip(),
        shared_object.strip(),
        execution_mode,
        symbol.strip(),
    )


def parse_perf_report_file(parsed_report_path):
    """
    Parses a performance report file and returns the parsed data as a pandas DataFrame.

    Args:
        parsed_report_path (str): The path to the parsed report file.

    Returns:
        pandas.DataFrame: The parsed data as a DataFrame with the following columns:
            - overhead_cycles: The number of overhead cycles.
            - overhead_insns: The number of overhead instructions.
            - command: The command associated with the performance data.
            - shared_object: The shared object associated with the performance data.
            - execution_mode: The execution mode associated with the performance data.
            - symbol: The symbol associated with the performance data.
    """
    lr = []
    with open(parsed_report_path, "r") as file:
        lines = file.readlines()
        for line in lines:
            if "......." in line:
                a = line.strip().removeprefix("#").split()
                for x in a:
                    if not lr:
                        l = line.index(x)
                        lr.append((l, l + len(x)))
                    else:
                        l = line.index(x, lr[-1][1])
                        lr.append((l, l + len(x)))
                break

    content = [l for l in lines if not l.startswith("#") and l.strip() != ""]

    if content is None:
        logger.info("content is None")
        return None

    data = [parse_one_line(l, lr) for l in content]

    data = [d for d in data if d is not None]

    logger.info("Successfully parsed data")
    logger.info("parsed data length: " + str(len(data)))

    return pd.DataFrame(
        data,
        columns=[
            "overhead_cycles",
            "overhead_insns",
            "command",
            "shared_object",
            "execution_mode",
            "symbol",
        ],
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pipad.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0bpipad.proto\x12\x04pipa"\xb3\r\n\rDeployRequest\x12\x10\n\x08workload\x18\x01 \x01(\t\x12\x14\n\x0ctransactions\x18\x02 \x01(\x04\x12\x12\n\nthroughput\x18\x03 \x01(\x01\x12\x14\n\x0cused_threads\x18\x04 \x03(\r\x12\x10\n\x08run_time\x18\x05 \x01(\x01\x12\x0e\n\x06\x63ycles\x18\x06 \x01(\x04\x12\x14\n\x0cinstructions\x18\x07 \x01(\x04\x12\x19\n\x11\x63ycles_per_second\x18\x08 \x01(\x01\x12\x1f\n\x17instructions_per_second\x18\t \x01(\x01\x12\x0b\n\x03\x43PI\x18\n \x01(\x01\x12\x1b\n\x13\x63ycles_per_requests\x18\x0b \x01(\x01\x12\x13\n\x0bpath_length\x18\x0c \x01(\x01\x12\x19\n\x11\x63pu_frequency_mhz\x18\r \x01(\x01\x12\x14\n\x07\x63pu_usr\x18\x0e \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x63pu_nice\x18\x0f \x01(\x01H\x01\x88\x01\x01\x12\x14\n\x07\x63pu_sys\x18\x10 \x01(\x01H\x02\x88\x01\x01\x12\x17\n\ncpu_iowait\x18\x11 \x01(\x01H\x03\x88\x01\x01\x12\x16\n\tcpu_steal\x18\x12 \x01(\x01H\x04\x88\x01\x01\x12\x14\n\x07\x63pu_irq\x18\x13 \x01(\x01H\x05\x88\x01\x01\x12\x15\n\x08\x63pu_soft\x18\x14 \x01(\x01H\x06\x88\x01\x01\x12\x16\n\tcpu_guest\x18\x15 \x01(\x01H\x07\x88\x01\x01\x12\x16\n\tcpu_gnice\x18\x16 \x01(\x01H\x08\x88\x01\x01\x12\x15\n\x08\x63pu_idle\x18\x17 \x01(\x01H\t\x88\x01\x01\x12\x15\n\x08\x63pu_util\x18\x18 \x01(\x01H\n\x88\x01\x01\x12\x16\n\tkbmemfree\x18\x19 \x01(\x04H\x0b\x88\x01\x01\x12\x14\n\x07kbavail\x18\x1a \x01(\x04H\x0c\x88\x01\x01\x12\x16\n\tkbmemused\x18\x1b \x01(\x04H\r\x88\x01\x01\x12\x1c\n\x0fpercent_memused\x18\x1c \x01(\x01H\x0e\x88\x01\x01\x12\x16\n\tkbbuffers\x18\x1d \x01(\x04H\x0f\x88\x01\x01\x12\x15\n\x08kbcached\x18\x1e \x01(\x04H\x10\x88\x01\x01\x12\x15\n\x08kbcommit\x18\x1f \x01(\x04H\x11\x88\x01\x01\x12\x1b\n\x0epercent_commit\x18  \x01(\x01H\x12\x88\x01\x01\x12\x15\n\x08kbactive\x18! \x01(\x04H\x13\x88\x01\x01\x12\x14\n\x07kbinact\x18" \x01(\x04H\x14\x88\x01\x01\x12\x14\n\x07kbdirty\x18# \x01(\x04H\x15\x88\x01\x01\x12\x15\n\x08kbanonpg\x18$ \x01(\x04H\x16\x88\x01\x01\x12\x13\n\x06kbslab\x18% \x01(\x04H\x17\x88\x01\x01\x12\x15\n\x08kbkstack\x18& \x01(\x04H\x18\x88\x01\x01\x12\x14\n\x07kbpgtbl\x18\' \x01(\x04H\x19\x88\x01\x01\x12\x15\n\x08kbvmused\x18( \x01(\x04H\x1a\x88\x01\x01\x12\x0b\n\x03\x64\x65v\x18) \x01(\t\x12\x10\n\x03tps\x18* \x01(\x01H\x1b\x88\x01\x01\x12\x12\n\x05rkB_s\x18+ \x01(\x01H\x1c\x88\x01\x01\x12\x12\n\x05wkB_s\x18, \x01(\x01H\x1d\x88\x01\x01\x12\x12\n\x05\x64kB_s\x18- \x01(\x01H\x1e\x88\x01\x01\x12\x14\n\x07\x61req_sz\x18. \x01(\x01H\x1f\x88\x01\x01\x12\x13\n\x06\x61qu_sz\x18/ \x01(\x01H \x88\x01\x01\x12\x17\n\ndisk_await\x18\x30 \x01(\x01H!\x88\x01\x01\x12\x1e\n\x11percent_disk_util\x18\x31 \x01(\x01H"\x88\x01\x01\x12\x15\n\rdata_location\x18\x32 \x01(\t\x12\x0f\n\x07hw_info\x18\x33 \x01(\t\x12\x0f\n\x07sw_info\x18\x34 \x01(\t\x12\x10\n\x08platform\x18\x35 \x01(\t\x12\x14\n\x07\x63omment\x18\x36 \x01(\tH#\x88\x01\x01\x12\x10\n\x08username\x18\x37 \x01(\tB\n\n\x08_cpu_usrB\x0b\n\t_cpu_niceB\n\n\x08_cpu_sysB\r\n\x0b_cpu_iowaitB\x0c\n\n_cpu_stealB\n\n\x08_cpu_irqB\x0b\n\t_cpu_softB\x0c\n\n_cpu_guestB\x0c\n\n_cpu_gniceB\x0b\n\t_cpu_idleB\x0b\n\t_cpu_utilB\x0c\n\n_kbmemfreeB\n\n\x08_kbavailB\x0c\n\n_kbmemusedB\x12\n\x10_percent_memusedB\x0c\n\n_kbbuffersB\x0b\n\t_kbcachedB\x0b\n\t_kbcommitB\x11\n\x0f_percent_commitB\x0b\n\t_kbactiveB\n\n\x08_kbinactB\n\n\x08_kbdirtyB\x0b\n\t_kbanonpgB\t\n\x07_kbslabB\x0b\n\t_kbkstackB\n\n\x08_kbpgtblB\x0b\n\t_kbvmusedB\x06\n\x04_tpsB\x08\n\x06_rkB_sB\x08\n\x06_wkB_sB\x08\n\x06_dkB_sB\n\n\x08_areq_szB\t\n\x07_aqu_szB\r\n\x0b_disk_awaitB\x14\n\x12_percent_disk_utilB\n\n\x08_comment"y\n\nDeployResp\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x10\n\x08username\x18\x02 \x01(\t\x12\x0c\n\x04time\x18\x03 \x01(\t\x12\x0c\n\x04hash\x18\x04 \x01(\t\x12\x17\n\x0fupload_datetime\x18\x05 \x01(\t\x12\x13\n\x0bstatus_code\x18\x06 \x01(\r2\xbb\x01\n\x05PIPAD\x12\x31\n\x06\x44\x65ploy\x12\x13.pipa.DeployRequest\x1a\x10.pipa.DeployResp"\x00\x12>\n\x11\x44\x65ployStreamReply\x12\x13.pipa.DeployRequest\x1a\x10.pipa.DeployResp"\x00\x30\x01\x12?\n\x10\x44\x65ployBidiStream\x12\x13.pipa.DeployRequest\x1a\x10.pipa.DeployResp"\x00(\x01\x30\x01\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "pipad_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_DEPLOYREQUEST"]._serialized_start = 22
    _globals["_DEPLOYREQUEST"]._serialized_end = 1737
    _globals["_DEPLOYRESP"]._serialized_start = 1739
    _globals["_DEPLOYRESP"]._serialized_end = 1860
    _globals["_PIPAD"]._serialized_start = 1863
    _globals["_PIPAD"]._serialized_end = 2050
# @@protoc_insertion_point(module_scope)



from pipa.common.cmd import run_command
import pytest


def test_run_command_success():
    # Test a successful command execution
    command = "echo 'Hello, World!'"
    output = run_command(command)
    assert output == "Hello, World!"


def test_run_command_failure():
    # Test a failed command execution
    command = "invalid_command"
    with pytest.raises(Exception):
        run_command(command)


def test_run_command_with_log():
    # Test command execution with logging enabled
    command = "echo 'Hello, World!'"
    output = run_command(command, log=True)
    assert output == "Hello, World!"


def test_run_command_with_custom_cwd():
    # Test command execution with a custom working directory
    command = "pwd"
    cwd = "/tmp"
    output = run_command(command, cwd=cwd)
    assert output == "/tmp"


if __name__ == "__main__":
    pytest.main([__file__])



"""
pipad_client.py

send performance data to server:

```python
from pipa.service.pipad.pipad_client import PIPADClient
import pipa.service.pipad.pipad_pb2 as pipadlib

client = PIPADClient(port="50051", address="127.0.0.1")
client.deploy(data=pipadlib.DeployRequest(
    transactions=3423523,
    throughput=251.57342,
    used_threads=[36,44],
    run_time=322.4164,
    cycles=1232142629,
    instructions=40847985,
    cycles_per_second=40636919.1644998,
    instructions_per_second=135971.3205,
    CPI=0.29867620232,
    cycles_per_requests=1613.3812311,
    path_length=540178.22362178,
    cpu_frequency_mhz=3435.1,
    cpu_usr=78.33,
    cpu_nice=0.0,
    cpu_sys=0.16,
    cpu_iowait=0.0,
    cpu_steal=0.0,
    cpu_irq=0.13,
    cpu_soft=0.0,
    cpu_guest=0.0,
    cpu_gnice=0.0,
    cpu_idle=0.32,
    cpu_util=99.68,
    kbmemfree=782348669,
    kbavail=7847456900,
    kbmemused=281231,
    percent_memused=0.2,
    kbbuffers=2631,
    kbcached=4323857,
    kbcommit=345630,
    percent_commit=0.43,
    kbactive=15324194,
    kbinact=5223120,
    kbdirty=456,
    kbanonpg=2102529,
    kbslab=556534,
    kbkstack=18573,
    kbpgtbl=39314,
    kbvmused=454329,
    tps=0.75,
    rkB_s=0.0,
    wkB_s=30.27,
    dkB_s=0.0,
    areq_sz=40.32,
    aqu_sz=0.01,
    disk_await=6.18,
    percent_disk_util=0.43,
    workload='rocksdb',
    data_location='xxx',
    dev='sdc',
    hw_info='1*1*1',
    sw_info='xxx',
    platform='xx',
    comment='',
))
```
"""

from typing import Optional
from pipa.common.logger import logger, stream_handler
import logging
import argparse
import grpc
from . import pipad_pb2 as pipadlib
from . import pipad_pb2_grpc as pipadgrpc


class PIPADClient:
    def __init__(
        self, port: Optional[int] = 5051, address: Optional[str] = "[::]"
    ) -> None:
        self._port = port
        self._address = address

    def deploy(self, data: pipadlib.DeployRequest) -> Optional[pipadlib.DeployResp]:
        """
        Deploys the performance data using the PIPADClient to server, will store them to specified database.

        Args:
            data (pipadlib.DeployRequest): The performance data to deploy.

        Returns:
            None
        """
        server = f"{self._address}:{self._port}"
        logger.info("try to deploy ...")
        try:
            with grpc.insecure_channel(server) as channel:
                stub = pipadgrpc.PIPADStub(channel)
                response: pipadlib.DeployResp = stub.Deploy(data)
            return response
        except Exception as e:
            logger.error(f"Client deploy received error: {e}")
            return None


if __name__ == "__main__":
    logger.setLevel(level=logging.DEBUG)
    stream_handler.setLevel(level=logging.DEBUG)
    argp = argparse.ArgumentParser()
    argp.add_argument(
        "-a", "--address", type=str, default="127.0.0.1", help="Specify address binding"
    )
    argp.add_argument(
        "-p", "--port", type=int, default=50051, help="Specify metrics port"
    )
    args = argp.parse_args()
    address = getattr(args, "address")
    port = getattr(args, "port")
    client = PIPADClient(port, address)
    resp = client.deploy(
        data=pipadlib.DeployRequest(
            transactions=7561946,
            throughput=25171.548361957342,
            used_threads=[36],
            run_time=300.416402331,
            cycles=1220032142629,
            instructions=4084798557821,
            cycles_per_second=4061136919.1644998,
            instructions_per_second=13597122281.3605,
            CPI=0.2986762077393152,
            cycles_per_requests=161338.38335119028,
            path_length=540178.2236769478,
            cpu_frequency_mhz=34545.1,
            cpu_usr=99.39,
            cpu_nice=0.0,
            cpu_sys=0.16,
            cpu_iowait=0.0,
            cpu_steal=0.0,
            cpu_irq=0.13,
            cpu_soft=0.0,
            cpu_guest=0.0,
            cpu_gnice=0.0,
            cpu_idle=0.32,
            cpu_util=99.68,
            kbmemfree=783648669,
            kbavail=784724900,
            kbmemused=2875231,
            percent_memused=0.36,
            kbbuffers=268061,
            kbcached=4379857,
            kbcommit=3459630,
            percent_commit=0.43,
            kbactive=1530294,
            kbinact=5261120,
            kbdirty=456,
            kbanonpg=2102529,
            kbslab=556534,
            kbkstack=18573,
            kbpgtbl=39314,
            kbvmused=454329,
            tps=0.75,
            rkB_s=0.0,
            wkB_s=30.27,
            dkB_s=0.0,
            areq_sz=40.32,
            aqu_sz=0.01,
            disk_await=6.18,
            percent_disk_util=0.43,
            workload="rocksdb",
            data_location="/home/xyjiang/project/rocksdb-perf/data/randomread/static/purple-openEuler/t1-rocksdb7.9.2/perf-stat-sar-less/",
            dev="sdc",
            hw_info="1*1*1",
            sw_info="1 thread, RocksDB 7.9.2 build in release mode, debug_level=0, threads_num=1, db_bench with benchmark.sh",
            platform="SPR 4510",
            comment="",
            username="HiAll",
        )
    )
    if resp is not None:
        if resp.status_code == 200:
            logger.info(
                f"Message: {resp.message}, Username: {resp.username}, Hash: {resp.hash}, Time: {resp.time}, Datetime: {resp.upload_datetime}"
            )
        else:
            logger.warning(f"PIPAD server resp faild: {resp.message}")



import pytest
from unittest.mock import patch, mock_open, MagicMock
import os

from pipa.service.gengerate.common import (
    ask_number,
    quest_basic,
    write_title,
    load_yaml_config,
    opener,
)


@patch("questionary.text")
def test_ask_number_default(mock_text):
    mock_text.return_value.ask.return_value = ""
    assert ask_number("Test question", 10) == 10


@patch("questionary.text")
def test_ask_number_input(mock_text):
    mock_text.return_value.ask.return_value = "5"
    assert ask_number("Test question", 10) == 5


@patch("questionary.text")
def test_ask_number_invalid_input(mock_text):
    mock_text.return_value.ask.return_value = "invalid"
    with pytest.raises(SystemExit):
        ask_number("Test question", 10)


@patch("questionary.text")
@patch("questionary.select")
@patch("os.path.exists")
@patch("os.makedirs")
def test_quest_basic_emon(mock_makedirs, mock_exists, mock_select, mock_text):
    mock_text.side_effect = [
        MagicMock(ask=lambda: ""),  # workspace
        MagicMock(ask=lambda: ""),  # freq_record
        MagicMock(ask=lambda: "{cycles,instructions}:S"),  # events_record
        MagicMock(ask=lambda: "/mnt/hdd/share/emon/system_health_monitor"),  # mpp
    ]
    mock_select.side_effect = [
        MagicMock(ask=lambda: "No"),  # annotete
        MagicMock(ask=lambda: "emon"),  # stat
    ]
    mock_exists.return_value = False

    result = quest_basic()
    expected_result = {
        "workspace": "./",
        "freq_record": 999,
        "events_record": "{cycles,instructions}:S",
        "use_emon": True,
        "mpp": "/mnt/hdd/share/emon/system_health_monitor",
        "annotete": False,
    }
    assert result == expected_result


@patch("builtins.open", new_callable=mock_open)
def test_write_title(mock_file):
    file = mock_file()
    write_title(file)
    file.write.assert_called_once()
    assert (
        "# The script generated by PIPA-TREE is used to collect performance data."
        in file.write.call_args[0][0]
    )


@patch("builtins.open", new_callable=mock_open, read_data='workspace: "./"\n')
def test_load_yaml_config(mock_file):
    config = load_yaml_config("dummy_path.yaml")
    assert config == {"workspace": "./"}


@patch("os.open")
def test_opener(mock_open):
    mock_open.return_value = 3
    descriptor = opener("dummy_path", os.O_RDWR)
    mock_open.assert_called_once_with(path="dummy_path", flags=os.O_RDWR, mode=0o755)
    assert descriptor == 3


if __name__ == "__main__":
    pytest.main([__file__])



import pandas as pd
import seaborn as sns
import re
from pipa.common.cmd import run_command
from pipa.common.hardware.cpu import NUM_CORES_PHYSICAL
import multiprocessing


class SarData:
    def __init__(self, sar_string: str):
        """
        Initialize a SAR object with the given SAR string.

        Args:
            sar_string (str): The SAR string to parse.

        """
        self.sar_data: list[pd.DataFrame] = parse_sar_string(sar_string)

    @classmethod
    def init_with_sar_txt(cls, sar_txt_path: str):
        """
        Initializes the SarData object using a SAR text file.

        Args:
            sar_txt_path (str): Path to the SAR text file.

        Returns:
            SarData: Initialized SarData object.
        """
        with open(sar_txt_path, "r") as f:
            sar_content = f.readlines()
        return cls(sar_content)

    @classmethod
    def init_with_sar_bin(cls, sar_bin_path: str):
        """
        Initializes the SarData object using a SAR binary file.

        Args:
            sar_bin_path (str): Path to the SAR binary file.

        Returns:
            SarData: Initialized SarData object.
        """
        sar_content = parse_sar_bin_to_txt(sar_bin_path)
        return cls(sar_content)

    def filter_dataframe(self, df, data_type: str = "detail"):
        """
        Filters the given dataframe based on the specified data type.

        Parameters:
        - df: pandas.DataFrame
            The dataframe to be filtered.
        - data_type: str, optional
            The type of data to filter. Valid values are "detail", "raw", and "average".
            Defaults to "detail".

        Returns:
        - pandas.DataFrame
            The filtered dataframe.

        Raises:
        - ValueError: If an invalid data type is provided.
        """
        match data_type:
            case "detail":
                return df[df["timestamp"] != "Average:"]
            case "raw":
                return df
            case "average":
                return df[df["timestamp"] == "Average:"]
            case _:
                raise ValueError("Invalid type")

    def get_CPU_utilization(self, data_type: str = "detail"):
        """
        Retrieves the CPU utilization data from the SAR data.

        Args:
            data_type (str, optional): The type of CPU utilization data to retrieve. Defaults to "detail".
            Valid values are "detail", "raw", and "average".

        Returns:
            DataFrame: The filtered DataFrame containing the CPU utilization data.
        """
        util = self.filter_dataframe(self.sar_data[0], data_type).astype(
            {
                r"%usr": "float64",
                r"%nice": "float64",
                r"%sys": "float64",
                r"%iowait": "float64",
                r"%irq": "float64",
                r"%soft": "float64",
                r"%steal": "float64",
                r"%guest": "float64",
                r"%gnice": "float64",
                r"%idle": "float64",
            }
        )
        util[r"%util"] = 100 - util[r"%idle"]
        return util

    def get_CPU_util_avg_by_threads(self, threads: list = None):
        """
        Retrieves the average CPU utilization detailed data for the specified threads.

        Args:
            threads (list): List of CPU threads to retrieve the utilization data for.
                            If None, retrieves the utilization data for all threads. Defaults to None.

        Returns:
            DataFrame: The filtered DataFrame containing the CPU utilization data for the specified threads.
        """
        util = self.get_CPU_utilization("average")
        util_threads = (
            util[util["CPU"].isin([str(t) for t in threads])]
            if threads
            else util[util["CPU"] == "all"]
        )
        return util_threads

    def get_CPU_util_avg_summary(self, threads: list = None):
        """
        Retrieves the average CPU utilization summary for the specified threads.

        Args:
            threads (list, optional): List of CPU threads to retrieve the utilization summary for. If None, retrieves
                                      the summary for all threads. Defaults to None.

        Returns:
            dict: A dictionary containing the average CPU utilization summary.
        """
        util_threads = self.get_CPU_util_avg_by_threads(threads)
        return util_threads.drop(columns=["timestamp", "CPU"]).mean().to_dict()

    def plot_CPU_util_time(self, threads: list = None):
        """
        Plots the CPU utilization over time.

        Args:
            threads (list, optional): List of CPU threads to plot. If None, plots the utilization for all threads.
                                      Defaults to None.
        """
        df = self.get_CPU_utilization()
        df = (
            df[df["CPU"].isin([str(t) for t in threads])]
            if threads
            else df.query("CPU=='all'")
        )
        df = trans_time_to_seconds(df)

        if threads and len(threads) > 1:
            sns.lineplot(data=df, x="timestamp", y=r"%util", hue="CPU")
        else:
            sns.lineplot(data=df, x="timestamp", y=r"%util")

    def get_CPU_frequency(self, data_type: str = "detail"):
        """
        Returns the CPU frequency data.

        Args:
            data_type (str, optional): The type of CPU freqency data to retrieve. Defaults to "detail".

        Returns:
            pd.DataFrame: Dataframe containing the CPU frequency data.
        """
        return self.filter_dataframe(self.sar_data[31], data_type).astype(
            {"MHz": "float64"}
        )

    def plot_CPU_freq_time(self, threads: list = None):
        """
        Plots the CPU frequency over time.

        Args:
            threads (list, optional): List of CPU threads to plot. If None, plots the frequency for all threads.
                                      Defaults to None.
        """
        df = self.get_CPU_frequency()

        sns.set_theme(style="darkgrid", rc={"figure.figsize": (15, 8)})

        df = (
            df[df["CPU"].isin([str(t) for t in threads])]
            if threads
            else df.query("CPU=='all'")
        )
        df = trans_time_to_seconds(df)

        if threads and len(threads) > 1:
            sns.lineplot(
                data=df,
                x="timestamp",
                y="MHz",
                hue="CPU",
            )
        else:
            sns.lineplot(data=df, x="timestamp", y="MHz")

    def get_cpu_freq_avg(self, threads: list = None):
        """
        Returns the average CPU frequency data for the specified threads.

        Args:
            threads (list, optional): List of CPU threads to retrieve the frequency data for. If None, retrieves the
                                      frequency data for all threads. Defaults to None.

        Returns:
            dict: A dictionary containing the average CPU frequency data for the specified threads.
        """
        df = self.get_CPU_frequency("average")
        df = df[df["CPU"].isin([str(t) for t in threads])] if threads else df
        if df.empty:
            return {"cpu_frequency_mhz": 0}
        return {"cpu_frequency_mhz": df["MHz"].mean()}

    def get_CPU_util_freq(self, data_type: str = "detail"):
        """
        Returns the CPU utilization and frequency data.

        Args:
            data_type (str, optional): The type of CPU utilization and frequency data to retrieve. Defaults to "detail".

        Returns:
            pd.DataFrame: Dataframe containing the CPU utilization and frequency data.
        """
        util, freq = self.get_CPU_utilization(data_type), self.get_CPU_frequency(
            data_type
        )
        return pd.merge(util, freq, on=["timestamp", "CPU"])

    def get_memory_usage(self, data_type: str = "detail"):
        """
        Returns the memory usage data.

        Args:
            data_type (str, optional): The type of memory usage data to retrieve. Defaults to "detail".

        Returns:
            pd.DataFrame: Dataframe containing the memory usage data.
        """
        return self.filter_dataframe(self.sar_data[6], data_type).astype(
            {
                "kbmemfree": int,
                "kbavail": int,
                "kbmemused": int,
                r"%memused": float,
                "kbbuffers": int,
                "kbcached": int,
                "kbcommit": int,
                r"%commit": float,
                "kbactive": int,
                "kbinact": int,
                "kbdirty": int,
                "kbanonpg": int,
                "kbslab": int,
                "kbkstack": int,
                "kbpgtbl": int,
                "kbvmused": int,
            }
        )

    def get_memory_usage_avg(self):
        """
        Returns the average memory usage data.

        Returns:
            dict: A dictionary containing the average memory usage data.
        """
        return (
            self.get_memory_usage("average")
            .drop(columns=["timestamp"])
            .to_dict(orient="records")[0]
        )

    def plot_memory_usage(self):
        """
        Plots the memory usage over time.
        """
        df = self.get_memory_usage()
        df = trans_time_to_seconds(df)
        sns.lineplot(data=df, x="timestamp", y=r"%memused")

    def get_disk_usage(self, dev: str = None, data_type: str = "detail"):
        """
        Returns the disk usage data.

        Args:
            data_type (str, optional): The type of disk usage data to retrieve. Defaults to "detail".

        Returns:
            pd.DataFrame: Dataframe containing the disk usage data.
        """
        df = self.filter_dataframe(self.sar_data[11], data_type).astype(
            {
                "tps": "float64",
                r"rkB/s": "float64",
                r"wkB/s": "float64",
                r"dkB/s": "float64",
                "areq-sz": "float64",
                "aqu-sz": "float64",
                "await": "float64",
                r"%util": "float64",
            }
        )
        return df[df["DEV"] == dev] if dev else df

    def get_disk_usage_avg(self, dev: str = None):
        """
        Returns the average disk usage data.

        Args:
            dev (str, optional): The disk device to retrieve the average data for. Defaults to None.

        Returns:
            list: A list of dictionaries containing the average disk usage data for each device.
            if dev is specified, returns a single dictionary.
        """
        disk_usage_avg = (
            self.get_disk_usage(dev, "average")
            .drop(columns=["timestamp"])
            .rename(columns={"%util": "%disk_util", "await": "disk_await"})
            .to_dict(orient="records")
        )
        return disk_usage_avg[0] if dev else disk_usage_avg

    def plot_disk_usage(self, dev: str = None, metrics="tps"):
        """
        Plots the disk tps over time.
        """
        df = self.get_disk_usage()
        df = trans_time_to_seconds(df).query(f"DEV=='{dev}'") if dev else df
        if dev:
            sns.lineplot(data=df, x="timestamp", y=metrics)
        else:
            sns.lineplot(data=df, x="timestamp", y=metrics, hue="DEV")


def parse_sar_bin_to_txt(sar_bin_path: str):
    """
    Parses the SAR binary file into a list of lines.

    Args:
        sar_bin_path (str): Path to the SAR binary file.

    Returns:
        list: List of lines in the SAR binary file.
    """
    sar_lines = run_command(f"LC_ALL='C' sar -A -f {sar_bin_path}").split("\n")
    return sar_lines


def split_sar_block(sar_lines: list):
    """
    Splits the SAR block into individual blocks by '\n'.

    Args:
        sar_lines (list): List of SAR output lines.

    Returns:
        list: List of individual SAR blocks.
    """
    sar_lines = [l.strip() for l in sar_lines]
    return [
        list(filter(None, p.split("\n"))) for p in "\n".join(sar_lines).split("\n\n")
    ]


def trans_time_to_seconds(df):
    """
    Transforms the timestamp column in the given DataFrame to seconds.

    Args:
        df (pandas.DataFrame): The DataFrame containing the timestamp column.

    Returns:
        pandas.DataFrame: The DataFrame with the timestamp column transformed to seconds.
    """
    df["timestamp"] = pd.to_datetime(df["timestamp"], format="%H:%M:%S")
    df["timestamp"] -= df.loc[:, "timestamp"].iloc[0]
    df["timestamp"] = df["timestamp"].dt.total_seconds()
    return df


def merge_one_line(sar_line: str) -> list:
    """
    Merge a single line of SAR data into a list.

    Args:
        sar_line (str): The SAR data line to be merged.

    Returns:
        list: The merged SAR data as a list.
    """
    sar_line = sar_line.split()
    if sar_line[1] in ["AM", "PM"]:
        sar_line.pop(1)
    return sar_line


def add_post_fix(sar_line: list, len_columns: int):
    """
    Adds post-fix to the given SAR line to match the specified number of columns.

    Args:
        sar_line (list): The SAR line to add post-fix to.
        len_columns (int): The desired number of columns.

    Returns:
        list: The SAR line with post-fix added to match the specified number of columns.
    """
    while len(sar_line) < len_columns:
        sar_line.append("")
    if len(sar_line) > len_columns:
        sar_line[len_columns - 1] += " ".join(sar_line[len_columns:])
    return sar_line[:len_columns]


def process_subtable(
    sar_columns: list, sar_blocks: list, processes_num=min(12, NUM_CORES_PHYSICAL)
):
    """
    Process the subtable data by merging lines and adding post-fixes.

    Args:
        sar_columns (list): List of SAR columns.
        sar_blocks (list): List of SAR blocks.
        processes_num (int, optional): Number of processes to use for parallel processing.
            Defaults to the minimum of 12 and the number of physical CPU cores.

    Returns:
        list: List of processed subtable data.

    """
    if len(sar_blocks) <= 10**6 or processes_num <= 1:
        # if the number of lines is less than 1e6, use single process
        return [add_post_fix(merge_one_line(x), len(sar_columns)) for x in sar_blocks]
    pool = multiprocessing.Pool(processes=processes_num)
    merged_lines = pool.map(merge_one_line, sar_blocks)
    res = pool.starmap(
        add_post_fix,
        zip(merged_lines, [len(sar_columns) for _ in range(len(merged_lines))]),
    )
    pool.close()
    pool.join()
    return res


def sar_to_df(sar_blocks: list):
    """
    Convert SAR blocks to a pandas DataFrame.

    Args:
        sar_blocks (list): A list of SAR blocks.

    Returns:
        pandas.DataFrame: A DataFrame containing the processed SAR data.

    """
    while sar_blocks[0] == "":
        sar_blocks = sar_blocks[1:]

    time_pattern = r"\d{2}:\d{2}:\d{2}"
    sar_columns = sar_blocks[0].split()
    if re.match(time_pattern, sar_columns[0]):
        sar_columns = ["timestamp"] + sar_columns[1:]
    return pd.DataFrame(
        process_subtable(sar_columns, sar_blocks[1:]),
        columns=sar_columns,
    )


def parse_sar_bin(sar_bin_path: str):
    """
    Parses the SAR binary file and returns a list of dataframes.

    Args:
        sar_bin_path (str): The path to the SAR binary file.

    Returns:
        List[pd.DataFrame]: A list of dataframes containing the parsed SAR data.
    """
    sar_content = parse_sar_bin_to_txt(sar_bin_path)
    return parse_sar_string(sar_content)


def parse_sar_txt(sar_txt_path: str):
    """
    Parses the SAR text file and returns a list of dataframes.

    Args:
        sar_txt_path (str): The path to the SAR text file.

    Returns:
        List[pd.DataFrame]: A list of dataframes containing the parsed SAR data.
    """
    with open(sar_txt_path, "r") as f:
        sar_content = f.readlines()
    return parse_sar_string(sar_content)


def parse_sar_string(sar_string: str):
    """
    Parses the SAR string and returns a list of dataframes.

    Args:
        sar_string (str): The string containing the SAR data.

    Returns:
        List[pd.DataFrame]: A list of dataframes containing the parsed SAR data.
    """
    sar_data = split_sar_block(sar_string)[1:]
    a = [sar_to_df(d) for d in sar_data]
    l = 0
    res = []
    while l < len(a):  # merge dataframes with the same columns, use two pointers
        r = l + 1
        while r < len(a) and a[r].columns.equals(a[l].columns):
            r += 1
        res.append(pd.concat(a[l:r], axis=0))
        l = r
    return res



import pytest
from unittest.mock import mock_open, patch
import pandas as pd
from pandas.testing import assert_frame_equal
from io import StringIO
from pipa.parser.perf_stat import PerfStatData


# Fixture for sample data
@pytest.fixture
def sample_data():
    return """
# started on Mon Jul 15 23:53:10 2024

1.001060857,CPU0,45892724,,cycles,1037560862,100.00,,
1.001060857,CPU1,31607000,,cycles,1037631576,100.00,,
1.001060857,CPU2,15936152,,cycles,1037681348,100.00,,
1.001060857,CPU3,7316201,,cycles,1037628759,100.00,,
1.001060857,CPU4,28015727,,cycles,1037603520,100.00,,
1.001060857,CPU5,37519594,,cycles,1037584024,100.00,,
1.001060857,CPU6,346613976,,cycles,1037519925,100.00,,
"""


# Fixture for expected DataFrame
@pytest.fixture
def expected_df():
    return pd.DataFrame(
        {
            "timestamp": [1.001060857] * 7,
            "cpu_id": list(range(7)),
            "value": [
                45892724,
                31607000,
                15936152,
                7316201,
                28015727,
                37519594,
                346613976,
            ],
            "unit": ["nan"] * 7,
            "metric_type": ["cycles"] * 7,
            "run_time(ns)": [
                1037560862,
                1037631576,
                1037681348,
                1037628759,
                1037603520,
                1037584024,
                1037519925,
            ],
            "run_percentage": [
                100.00,
                100.00,
                100.00,
                100.00,
                100.00,
                100.00,
                100.00,
            ],
            "opt_value": [float("nan")] * 7,
            "opt_unit_metric": ["nan"] * 7,
        }
    )


@patch("builtins.open", new_callable=mock_open)
def test_parse_perf_stat_file(mock_file, sample_data, expected_df):
    mock_file.return_value = StringIO(sample_data.strip())

    result_df = PerfStatData.parse_perf_stat_file("dummy_path")

    result_df.replace("", float("nan"), inplace=True)

    # Compare the result with the expected DataFrame
    assert_frame_equal(result_df, expected_df, check_like=True)


if __name__ == "__main__":
    pytest.main([__file__])
